from adios2.adios import Adios

import adios2.bindings as bindings

import unittest
import numpy as np


class TestOperator(unittest.TestCase):
    def test_operator_basic(self):
        adios = Adios()
        op1 = adios.define_operator("noop", "null")
        with adios.declare_io("BPWriter") as io:
            temps = io.define_variable("temps", np.empty([4], dtype=np.int64))
            temps.add_operation(op1)
            with io.open("pythontestvariable.bp", bindings.Mode.Write) as engine:
                temps_measures = np.array([35, 40, 30, 45], dtype=np.int64)
                engine.put(temps, temps_measures)

        op2 = adios.define_operator("noop2", "null")
        with adios.declare_io("BPReader") as reader:
            with reader.open("pythontestvariable.bp", bindings.Mode.Read) as engine:
                engine.begin_step()
                temps = reader.inquire_variable("temps")
                temps.add_operation(op2)
                engine.end_step()

    def test_operator_params(self):
        adios = Adios()
        op = adios.define_operator("noop", "null")
        op.set_parameter("speed", "best")
        op.get_parameters()
        self.assertTrue("speed" in op.get_parameters())
        self.assertEqual(op.get_parameters()["speed"], "best")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import adios2  # pylint: disable=import-error
import argparse
import numpy as np  # pylint: disable=import-error
import matplotlib.pyplot as plt  # pylint: disable=import-error
import matplotlib.gridspec as gridspec  # pylint: disable=import-error
import decomp  # pylint: disable=import-error


def SetupArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--instream", "-i", help="Name of the input stream", required=True
    )
    parser.add_argument(
        "--outfile", "-o", help="Name of the output file", default="screen"
    )
    parser.add_argument("--varname", "-v", help="Name of variable read", default="U")
    parser.add_argument(
        "--nompi", "-nompi", help="ADIOS was installed without MPI", action="store_true"
    )
    parser.add_argument(
        "--displaysec",
        "-dsec",
        help="Float representing gap between plot window refresh",
        default=0.1,
    )
    parser.add_argument(
        "--nx",
        "-nx",
        help="Integer representing process decomposition in the x direction",
        default=1,
    )
    parser.add_argument(
        "--ny",
        "-ny",
        help="Integer representing process decomposition in the y direction",
        default=1,
    )
    parser.add_argument(
        "--nz",
        "-nz",
        help="Integer representing process decomposition in the z direction",
        default=1,
    )
    parser.add_argument(
        "--plane",
        "-plane",
        help="The 2D plane to be displayed/stored xy/yz/xz/all",
        default="yz",
    )
    args = parser.parse_args()

    args.displaysec = float(args.displaysec)
    args.nx = int(args.nx)
    args.ny = int(args.ny)
    args.nz = int(args.nz)

    if args.plane not in ("xz", "yz", "xy", "all"):
        raise ValueError("Input argument --plane must be one of xz/yz/xy/all")

    return args


def Plot2D(plane_direction, data, args, fullshape, step, fontsize):
    # Plotting part
    displaysec = args.displaysec
    gs = gridspec.GridSpec(1, 1)
    fig = plt.figure(1, figsize=(8, 8))
    ax = fig.add_subplot(gs[0, 0])
    colorax = ax.imshow(
        data,
        origin="lower",
        interpolation="quadric",
        extent=[0, fullshape[1], 0, fullshape[0]],
        cmap=plt.get_cmap("gist_ncar"),
    )
    cbar = fig.colorbar(colorax, orientation="horizontal")
    cbar.ax.tick_params(labelsize=fontsize - 4)

    for i in range(args.ny):
        y = fullshape[0] / args.ny * i
        ax.plot([0, fullshape[1]], [y, y], color="black")

    for i in range(args.nx):
        x = fullshape[1] / args.nx * i
        ax.plot([x, x], [0, fullshape[0]], color="black")

    ax.set_title(
        "{0}, {1} plane, step {2}".format(args.varname, plane_direction, step),
        fontsize=fontsize,
    )
    ax.set_xlabel(plane_direction[0], fontsize=fontsize)
    ax.set_ylabel(plane_direction[1], fontsize=fontsize)
    plt.tick_params(labelsize=fontsize - 8)
    plt.ion()
    if args.outfile == "screen":
        plt.show()
        plt.pause(displaysec)
    elif args.outfile.endswith(".bp"):
        if step == 0:
            global adios
            global ioWriter
            global var
            global writer
            adios = adios2.ADIOS(mpi.comm_app)
            ioWriter = adios.DeclareIO("VizOutput")
            var = ioWriter.DefineVariable(
                args.varname, data.shape, [0, 0], data.shape, adios2.ConstantDims, data
            )
            writer = ioWriter.Open(args.outfile, adios2.Mode.Write)

        writer.BeginStep()
        writer.Put(var, data, adios2.Mode.Sync)
        writer.EndStep()
    else:
        imgfile = args.outfile + "{0:0>5}".format(step) + "_" + plane_direction + ".png"
        fig.savefig(imgfile)

    plt.clf()


def read_data(args, fr, start_coord, size_dims):
    var1 = args.varname
    data = fr.read(var1, start_coord, size_dims)
    data = np.squeeze(data)
    return data


if __name__ == "__main__":
    # fontsize on plot
    fontsize = 24

    args = SetupArgs()
    #    print(args)

    # Setup up 2D communicators if MPI is installed
    mpi = decomp.MPISetup(args, 3)
    myrank = mpi.rank["app"]

    # Read the data from this object
    fr = adios2.open(args.instream, "r", mpi.comm_app, "adios2.xml", "SimulationOutput")
    #    vars_info = fr.availablevariables()

    # Get the ADIOS selections -- equally partition the data if parallelization is requested

    # Read through the steps, one at a time
    plot_step = 0
    for fr_step in fr:
        #        if fr_step.current_step()
        start, size, fullshape = mpi.Partition_3D_3D(fr, args)
        cur_step = fr_step.current_step()
        vars_info = fr.available_variables()
        #        print (vars_info)
        shape3_str = vars_info[args.varname]["Shape"].split(",")
        shape3 = list(map(int, shape3_str))
        sim_step = fr_step.read("step")

        if myrank == 0:
            print(
                "GS Plot step {0} processing simulation output step {1} or computation"
                "step {2}".format(plot_step, cur_step, sim_step),
                flush=True,
            )
        #            if cur_step == 0:
        #                print("Variable" + pdfvar + " shape is {" + vars_info[pdfvar]["Shape"]+"}")

        if args.plane in ("xy", "all"):
            data = read_data(
                args, fr_step, [0, 0, int(shape3[2] / 2)], [shape3[0], shape3[1], 1]
            )
            Plot2D("xy", data, args, fullshape, sim_step, fontsize)

        if args.plane in ("xz", "all"):
            data = read_data(
                args, fr_step, [0, int(shape3[1] / 2), 0], [shape3[0], 1, shape3[2]]
            )
            Plot2D("xz", data, args, fullshape, sim_step, fontsize)

        if args.plane in ("yz", "all"):
            data = read_data(
                args, fr_step, [int(shape3[0] / 2), 0, 0], [1, shape3[1], shape3[2]]
            )
            Plot2D("yz", data, args, fullshape, sim_step, fontsize)
        plot_step = plot_step + 1

    fr.close()



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# bpWriter-bindings.py : Low-level Python API example
#  Created on: Feb 2, 2017
#      Authors: William F Godoy godoywf@ornl.gov
#               Norbert Podhorszki pnorbert@ornl.gov
#
# adios2.bindings gives access to the low level classes that are compiled from the C++ classes
# Only necessary to use this when the adios2 python API is insufficient for some reason
# This example doesn't use anything that is missing from the python API though.

from mpi4py import MPI
import numpy
import adios2.bindings as adios2

# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# User data
myArray = numpy.array([0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])
myArray = myArray + (rank * 10)
Nx = myArray.size

# ADIOS MPI Communicator
adios = adios2.ADIOS(comm)

# ADIOS IO
bpIO = adios.DeclareIO("BPFile_N2N")
bpIO.SetEngine("BPFile")
# bpIO.SetParameters( {"Threads" : "2", "ProfileUnits" : "Microseconds",
# "InitialBufferSize" : "17Kb"} )
bpIOParams = {}
bpIOParams["Threads"] = "2"
bpIOParams["ProfileUnits"] = "Microseconds"
bpIOParams["InitialBufferSize"] = "17Kb"
bpIO.SetParameters(bpIOParams)

fileID = bpIO.AddTransport("File", {"Library": "fstream"})

# ADIOS Variable name, shape, start, offset, constant dims
ioArray = bpIO.DefineVariable(
    "bpArray", myArray, [size * Nx], [rank * Nx], [Nx], adios2.ConstantDims
)

varNx = bpIO.DefineVariable("Nx", numpy.array(Nx))  # type is derived from numpy array type
bpIO.DefineAttribute("size", Nx, "bpArray")
bpIO.DefineAttribute("dimensions", ["Nx"], "bpArray")

# ADIOS Engine
bpFileWriter = bpIO.Open("bpWriter-py-bindings.bp", adios2.Mode.Write)
bpFileWriter.BeginStep()
bpFileWriter.Put(ioArray, myArray, adios2.Mode.Sync)
bpFileWriter.Put(varNx, numpy.array(Nx), adios2.Mode.Sync)
bpFileWriter.EndStep()
bpFileWriter.Close()

# Read content:
# bpls -la bpWriter-py-bindings.bp
# bpls -la bpWriter-py-bindings.bp -d bpArray -n 10
# bpls -la bpWriter-py-bindings.bp -d bpArray -n 10 -D



import argparse
from paraview.simple import *

from paraview import print_info

# ------------------------------------------------------------------------------
# Catalyst options
from paraview import catalyst

options = catalyst.Options()
options.GlobalTrigger = "TimeStep"
options.EnableCatalystLive = 1
options.CatalystLiveTrigger = "TimeStep"
options.ExtractsOutputDirectory = "/tmp"
# options.ExtractsOutputDirectory = '.'

DISABLE_EXTRACTOR = False

# setup and returns the view
def SetupRenderView():
    # Create a new 'Render View'
    view = CreateView("RenderView")

    camera = GetActiveCamera()
    camera.Azimuth(45)
    camera.Elevation(45)

    SetActiveView(None)

    # create new layout object 'Layout #1'
    layout1 = CreateLayout(name="Layout #1")
    layout1.AssignView(0, view)
    layout1.SetSize(824, 656)

    # restore active view
    SetActiveView(view)
    return view


# A different proxy type is used depending on whether you're using
# catalyst, or doing post hoc visualization
# this returns the catalyst producer
def SetupCatalystProducer():
    producer = TrivialProducer(registrationName="fides")
    return producer


# this returns the fides proxy which should be used for post hoc vis
# i.e., you're reading .bp files
def SetupFidesReader(json, bp, sst):
    if json is None:
        # in this case the bp file must contain the Fides attributes
        fides = FidesReader(StreamSteps=1, FileName=bp)
        return fides

    fides = FidesJSONReader(StreamSteps=1, FileName=json)
    if sst:
        fides.DataSourceEngines = ["source", "SST"]
    # 'source' is the name of the ADIOS data source in the JSON data model
    fides.DataSourcePath = ["source", bp]
    # required to update the fides reader
    fides.UpdatePipelineInformation()
    return fides


# takes in a producer and view and sets up the visualization pipeline
def SetupVisPipeline(producer, view):
    view.ResetCamera()

    # get color transfer function/color map for 'U'
    uLUT = GetColorTransferFunction("V")
    uLUT.AutomaticRescaleRangeMode = "Clamp and update every timestep"
    uLUT.RescaleOnVisibilityChange = 1

    contour1 = Contour(registrationName="Contour1", Input=producer)
    contour1.ContourBy = ["POINTS", "V"]
    contour1.Isosurfaces = [0.1, 0.3, 0.5, 0.7]
    contour1.PointMergeMethod = "Uniform Binning"

    contour1Display = Show(contour1, view, "GeometryRepresentation")

    contour1Display.Representation = "Surface"
    contour1Display.ColorArrayName = ["POINTS", "V"]

    Hide(producer, view)

    # trace defaults for the display properties.
    contour1Display.Representation = "Surface"
    contour1Display.ColorArrayName = ["POINTS", "V"]
    contour1Display.SetScaleArray = ["POINTS", "V"]
    contour1Display.ScaleTransferFunction = "PiecewiseFunction"
    contour1Display.LookupTable = uLUT

    uLUTColorBar = GetScalarBar(uLUT, view)
    uLUTColorBar.Title = "V"
    uLUTColorBar.ComponentTitle = ""

    # show color legend
    contour1Display.SetScalarBarVisibility(view, True)
    return contour1, contour1Display


# sets up an extractor for writing out PNG images
def SetupExtractor(view):
    # create extractor
    pNG1 = CreateExtractor("PNG", view, registrationName="PNG1")
    # trace defaults for the extractor.
    pNG1.Trigger = "TimeStep"

    # init the 'PNG' selected for 'Writer'
    pNG1.Writer.FileName = "output_{timestep:06d}.png"
    pNG1.Writer.ImageResolution = [800, 800]
    pNG1.Writer.Format = "PNG"


# Catalyst uses this to update the pipeline and print out some info on each time step
# you don't need to call this directly in your script; ParaView Catalyst will call it for you
def catalyst_execute(info):
    print_info("in '%s::catalyst_execute'", __name__)
    global pipeline, display
    pipeline.UpdatePipeline()
    display.RescaleTransferFunctionToDataRange()

    print_info("executing (cycle={}, time={})".format(info.cycle, info.time))
    print_info("U-range: {}".format(producer.PointData["U"].GetRange(0)))
    print_info("V-range: {}".format(producer.PointData["V"].GetRange(0)))


def ParseArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-j",
        "--json_filename",
        help="path to Fides JSON file",
        type=str,
        required=False,
    )
    parser.add_argument(
        "-b", "--bp_filename", help="path to bp file", type=str, required=True
    )
    parser.add_argument("--staging", help="use SST engine", action="store_true")
    args = parser.parse_args()
    return args


def StreamingVis(args):
    # adios/fides step status
    NotReady = 1
    EndOfStream = 2

    # setup the reader, view, pipeline
    fides = SetupFidesReader(args.json_filename, args.bp_filename, args.staging)
    view = SetupRenderView()

    step = 0
    while True:
        status = NotReady
        while status == NotReady:
            # must call PrepareNextStep to get Fides ready to read the
            # next step
            fides.PrepareNextStep()
            fides.UpdatePipelineInformation()
            status = fides.NextStepStatus
        if status == EndOfStream:
            # done reading the file
            return
        if step == 0:
            # set up the pipeline on the first step
            pipeline, display = SetupVisPipeline(fides, view)

        # need to update the pipeline and then save the output
        pipeline.UpdatePipeline()
        display.RescaleTransferFunctionToDataRange()
        output = f"output-{step:05d}.png"
        SaveScreenshot(output, view, ImageResolution=[800, 800])
        step += 1


# ------------------------------------------------------------------------------
if __name__ == "__main__":
    print("in __main__()")
    args = ParseArgs()
    StreamingVis(args)
else:
    # in this case we're running from Catalyst
    view = SetupRenderView()
    producer = SetupCatalystProducer()
    pipeline, display = SetupVisPipeline(producer, view)

    # normally not needed, but a bug fix is in progress to fix an issue
    # when using ParaView Live with extractors
    if not DISABLE_EXTRACTOR:
        SetupExtractor(view)



#!/usr/bin/env python3

# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# generate_pipeline.py
#
#   Author: Vicente Adolfo Bolea Sanchez <vicente.bolea@kitware.com>

from datetime import datetime
import argparse
import itertools
import requests
import time
import re
import urllib3
import git
import gitdb

# Remove annoying warning about insecure connection (self-signed cert).
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class skip_after_n_successes:
    def __init__(self, default_value, n):
        self.runs_max = n
        self.runs_current = 0
        self.default_value = default_value

    def __call__(self, fn, *args, **kwargs):
        if self.runs_current >= self.runs_max:
            return self.default_value

        ret = fn(*args, **kwargs)
        if ret:
            self.runs_current += 1
        return ret


def http_get_request(*args, **kwargs):
    kwargs["verify"] = False
    return requests.get(*args, **kwargs)


def request_as_list(url, *args, **kwargs):
    current_url = url
    body_json = []
    while current_url:
        response = http_get_request(current_url, *args, **kwargs)
        body_json += response.json()

        header = response.headers
        current_url = None
        if "link" in header:
            links = re.search(
                r'(?<=\<)([\S]*)(?=>; rel="next")', header["link"], flags=re.IGNORECASE
            )
            if links is not None:
                current_url = links.group(0)

    return body_json


def add_timestamp(branch):
    date_str = branch["commit"]["committed_date"]
    # We ignore the TZ since Gitlab/GitHub always reports in UTC
    branch["dt"] = int(datetime.strptime(date_str.split(".")[0], "%Y-%m-%dT%H:%M:%S").timestamp())
    return branch


def is_descendant_of_root(branch, root_commit, repo):
    if root_commit is None:
        return True
    try:
        commit_sha = branch["commit"]["id"]
        repo.commit(commit_sha)
        merge_base = repo.merge_base(args.root_commit, commit_sha)
        if len(merge_base) >= 1 and str(merge_base[0]) == args.root_commit:
            return True
    except gitdb.exc.BadObject:
        return False


def is_recent(branch):
    deadline_sec = int(time.time()) - (args.days * 86400)
    return branch["dt"] > deadline_sec


def has_no_status(branch):
    gh_commit_sha = branch["commit"]["id"]
    # Backported branches use the merge head
    if re.fullmatch(r"^pr\d+_.*$", branch["name"]):
        gh_commit_sha = branch["commit"]["parent_ids"][1]

    # Query GitHub for the status of this commit
    response = http_get_request(gh_url + "/commits/" + gh_commit_sha + "/status")
    if int(response.headers["x-ratelimit-remaining"]) <= 0:
        raise ConnectionError(response.json())

    commit = response.json()
    if commit is None or "sha" not in commit:
        return False

    for status in commit["statuses"]:
        if status["context"] == args.gh_context:
            return False

    return True


parser = argparse.ArgumentParser(
    prog="generate_pipeline.py", description="Generate Dynamic pipelines for Gitlab"
)
parser.add_argument(
    "-u",
    "--gl-url",
    required=True,
    help="Base URL for Gitlab remote. Ex: https://code.olcf.ornl.gov/",
)
parser.add_argument(
    "-n", "--gh-name", required=True, help="Full name of the GitHub project. Ex: ornladios/ADIOS2"
)
parser.add_argument(
    "-p", "--project_id", required=True, help="Gitlab internal project ID of the project."
)
parser.add_argument(
    "-c",
    "--gh-context",
    default="OLCF Frontier",
    help="Name of the status in GitHub (A.K.A context)",
)
parser.add_argument(
    "-d", "--days", type=int, default=1, help="How many days back to search for commits"
)
parser.add_argument(
    "-m", "--max", type=int, default=2, help="Maximum amount of pipelines computed"
)
parser.add_argument(
    "-r",
    "--root-commit",
    default=None,
    type=str,
    help="Template file of the pipeline `{branch}` will be substituted",
)
parser.add_argument(
    "-f",
    "--template_file",
    required=True,
    help="Template file of the pipeline `{branch}` will be substituted",
)
args = parser.parse_args()


gl_url = args.gl_url + "/api/v4/projects/" + str(args.project_id)
gh_url = "https://api.github.com/repos/" + args.gh_name

with open(args.template_file, "r") as fd:
    template_str = fd.read()

    repo = git.Repo(".", odbt=git.GitDB)

    branches = request_as_list(gl_url + "/repository/branches")
    branches = [b for b in branches if is_descendant_of_root(b, args.root_commit, repo)]
    branches = [add_timestamp(branch) for branch in branches]
    branches = [b for b in branches if is_recent(b)]
    branches = sorted(branches, key=lambda x: x["dt"])

    # Skip running (and return true) has_no_status after returning True args.max times.
    # We need this not to hog the Github Rest API draconian ratelimit.
    run_n_times = skip_after_n_successes(default_value=False, n=args.max)
    branches = [b for b in branches if run_n_times(has_no_status, b)]

    for branch in branches:
        print(template_str.format(branch=branch["name"], commit=branch["commit"]["id"]))



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# Test reader in Python that should have the same output
# as the test1.m Matlab reader
#
#
#  Created on: Jan 28, 2019
#      Author: Norbert Podhorszki, pnorbert@ornl.gov
#

import numpy
import adios2

adios = adios2.ADIOS()
io = adios.DeclareIO("reader")


# ADIOS2 high-level API for Reading
fr = io.Open("test1.bp", adios2.Mode.Read)


anote = io.InquireAttribute("anote")
adimnames = io.InquireAttribute("adimnames")
adims = io.InquireAttribute("adims")

dimNames = adimnames.DataString()
dims = adims.Data()

print("Info based on attributes in file:")
print("  Dimensions  {0}  {1}".format(dimNames[0], dimNames[1]))
print("               {0}     {1}".format(dims[0], dims[1]))

nrows = numpy.zeros(1, dtype=numpy.int64)
varRows = io.InquireVariable("nrows")
fr.Get(varRows, nrows)

ncols = numpy.zeros(1, dtype=numpy.int64)
varCols = io.InquireVariable("ncols")
fr.Get(varCols, ncols)

note = "_______________________________________"
varNote = io.InquireVariable("note")
fr.Get(varNote, note, adios2.Mode.Sync)

fr.PerformGets()

print("# of rows = {0}".format(nrows[0]))
print("# of cols = {0}".format(ncols[0]))
print("Note = {0}".format(note))
#
# inTemperatures = fr.read("temperature2D")
# print("temperature2d array size = " + str(inTemperatures.size))
#
# for row in inTemperatures:
#     print(''.join(['{:7}'.format(item) for item in row]))
#
fr.Close()



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# bpWriter.py : only works with MPI version
#  Created on: Feb 2, 2017
#      Authors: William F Godoy godoywf@ornl.gov
#               Norbert Podhorszki pnorbert@ornl.gov
#
# We use adios2.Adios and adios2.IO classes to set up the IO parameters.
# For default IO, it is sufficient to use the adios2.Stream class alone.

from mpi4py import MPI
import numpy
import adios2

# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# User data
myArray = numpy.array([0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])
myArray = myArray + (rank * 10)
Nx = myArray.size

# ADIOS MPI Communicator
adios = adios2.Adios(config_file=None, comm=comm)

# ADIOS IO
bpIO = adios.declare_io("BPFile_N2N")

bpIOParams = {}
bpIOParams["Threads"] = "2"
bpIOParams["ProfileUnits"] = "Microseconds"
bpIOParams["InitialBufferSize"] = "17Kb"
bpIO.set_parameters(bpIOParams)

bpIO.add_transport("File", {"Library": "fstream"})
bpIO.set_engine("BPFile")
a = bpIO.adios()

# ADIOS output stream
with adios2.Stream(bpIO, "bpWriter-py.bp", "w", comm) as fh:
    fh.write("bpArray", myArray, [size * Nx], [rank * Nx], [Nx])
    fh.write("Nx", Nx)
    fh.write_attribute("size", Nx, "bpArray")
    fh.write_attribute("dimensions", ["Nx"], "bpArray")

# Read content:
# bpls -la bpWriter-py.bp
# bpls -la bpWriter-py.bp -d bpArray -n 10
# bpls -la bpWriter-py.bp -d bpArray -n 10 -D



from adios2.adios import Adios
import unittest


class Test_adios(unittest.TestCase):
    def test_define_operator(self):
        adios = Adios()
        op = adios.define_operator("op", "null")
        self.assertNotEqual(op, None)

    def test_inquiry_operator(self):
        adios = Adios()
        op1 = adios.define_operator("op1", "null")
        op2 = adios.define_operator("op2", "null")
        op_x = adios.inquire_operator("op2")
        self.assertNotEqual(op1, op2)
        self.assertEqual(op2, op_x)
        self.assertNotEqual(op1, op_x)

    def test_declare_io(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        self.assertNotEqual(io, None)

    def test_at_io(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        reader = adios.declare_io("BPReader")
        x = adios.at_io("BPReader")
        self.assertNotEqual(io, reader)
        self.assertEqual(reader, x)
        self.assertNotEqual(io, x)

    def test_remove_io(self):
        adios = Adios()
        adios.declare_io("BPWriter")
        adios.remove_io("BPWriter")

    def test_remove_all_io(self):
        adios = Adios()
        adios.declare_io("BPWriter")
        adios.declare_io("BPReader")
        adios.remove_all_ios()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import os
import time
import argparse
import numpy  # pylint: disable=import-error
import adios2  # pylint: disable=import-error
import diagram  # pylint: disable=import-error


def run():
    parser = argparse.ArgumentParser(description=("Graph KdV equation"))
    parser.add_argument(
        "-f", "--filename", help=".bp filename", default="korteweg_de_vries.bp"
    )
    options = parser.parse_args()
    dgoptions = diagram.DOption()
    dgoptions.mode = "g"

    with adios2.open(options.filename, "r") as file_handle:
        for step in file_handle:
            points = step.read("u")
            # Adding these points helps us with whiplash;
            # otherwise the data is minmaxed on every iteration.
            # It does produce a visual artefact that I wish wasn't there!
            points = numpy.append(points, [3.0, -1.5])
            values = [None]
            dg = diagram.DGWrapper(dg_option=dgoptions, data=[points, values])
            os.system("cls" if os.name == "nt" else "clear")
            dg.show()
            time.sleep(0.1)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3

import argparse
import glob
from os.path import basename, exists, isdir

from adios2.bp5dbg import DumpIndexTable, DumpMetaData, DumpMetaMetaData

MetadataIndexTable = []
WriterMap = []
status = True


def SetupArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "FILE", help="Name of the input file (.bp, .bp/md.idx, " +
        ".bp/md.0 or .bp/data.XXX)")
    # parser.add_argument("--printdata", "-p",
    #                    help="Dump data of this variable as well", default="")
    parser.add_argument("--verbose", "-v",
                        help="More verbosity", action="count")
    parser.add_argument("--no-indextable", "-x",
                        help="Do not print index table md.idx",
                        action="store_true")
    parser.add_argument("--no-metadata", "-m",
                        help="Do not print metadata md.0",
                        action="store_true")
    parser.add_argument("--no-metametadata", "-M",
                        help="Do not print meta-metadata mmd.0",
                        action="store_true")
    parser.add_argument("--no-data", "-d",
                        help="Do not print data data.*", action="store_true")
    args = parser.parse_args()

    # default values
    args.idxFileName = ""
    args.dumpIdx = False
    args.metametadataFileName = ""
    args.dumpMetaMetadata = False
    args.metadataFileName = ""
    args.dumpMetadata = False
    args.dataFileName = ""
    args.dumpData = False

    # print("Verbosity = {0}".format(args.verbose))
    return args


def CheckFileName(args):
    if not exists(args.FILE):
        print("ERROR: File " + args.FILE + " does not exist", flush=True)
        exit(1)
    if isdir(args.FILE):
        if not args.no_indextable:
            args.idxFileName = args.FILE + "/" + "md.idx"
            args.dumpIdx = True
        if not args.no_metadata:
            args.metadataFileName = args.FILE + "/" + "md.[0-9]*"
            args.dumpMetadata = True
        if not args.no_metametadata:
            args.metametadataFileName = args.FILE + "/" + "mmd.[0-9]*"
            args.dumpMetaMetadata = True
        if not args.no_data:
            args.dataFileName = args.FILE + "/" + "data.[0-9]*"
            args.dumpData = True
        return

    name = basename(args.FILE)
    if name.startswith("data."):
        args.dataFileName = args.FILE
        args.dumpData = True

    elif name == "md.idx":
        args.idxFileName = args.FILE
        args.dumpIdx = True

    elif name.startswith("md."):
        args.metadataFileName = args.FILE
        args.dumpMetadata = True

    elif name.startswith("mmd."):
        args.metametadataFileName = args.FILE
        args.dumpMetaMetadata = True


def DumpIndexTableFile(args):
    global MetadataIndexTable
    global WriterMap
    global status
    indexFileList = glob.glob(args.idxFileName)
    if len(indexFileList) > 0:
        status, MetadataIndexTable, WriterMap = DumpIndexTable(
            indexFileList[0], True)
    else:
        print("There is  no BP% Index Table file as " + args.idxFileName)
        status = False
    return status


def DumpMetaMetadataFiles(args):
    global status
    mdFileList = glob.glob(args.metametadataFileName)
    if len(mdFileList) > 0:
        for fname in mdFileList:
            status = DumpMetaMetaData(fname)
    else:
        print("There are no BP% MetaMetadata files in   " +
              args.metametadataFileName)
        status = False
    return status


# xxx/md.X to xxx/md.idx
def GetIndexFileName(MDFileName):
    return MDFileName.rsplit('.', 1)[0] + ".idx"


def DumpMetadataFiles(args):
    global MetadataIndexTable
    global WriterMap
    global status
    mdFileList = glob.glob(args.metadataFileName)
    if len(mdFileList) > 0:
        if len(MetadataIndexTable) == 0:
            # need to parse index first
            IndexFileName = GetIndexFileName(mdFileList[0])
            status, MetadataIndexTable, WriterMap = DumpIndexTable(
                IndexFileName, False)

        if status:
            for fname in mdFileList:
                DumpMetaData(fname, MetadataIndexTable, WriterMap)
    else:
        print("There are no BP% Metadata files in   " + args.metadataFileName)
        status = False
    return status

# def DumpDataFiles(args):
#    dataFileList = glob.glob(args.dataFileName)
#    if len(dataFileList) > 0:
#        for fname in dataFileList:
#            DumpData(fname)
#    else:
#        print("There are no BP5 Data files in       " + args.dataFileName)


if __name__ == "__main__":

    args = SetupArgs()
    CheckFileName(args)
    # print(args)

    if args.dumpIdx:
        status = DumpIndexTableFile(args)

    if args.dumpMetaMetadata and status:
        status = DumpMetaMetadataFiles(args)

    if args.dumpMetadata and status:
        status = DumpMetadataFiles(args)

#    if args.dumpData:
#        DumpDataFiles(args)



"""License:
  Distributed under the OSI-approved Apache License, Version 2.0.  See
  accompanying file Copyright.txt for details.
"""

from functools import singledispatchmethod
from adios2 import Stream, IO


# pylint: disable=W0221
class FileReader(Stream):
    """High level implementation of the FileReader class for read Random access mode"""

    def __repr__(self):
        return f"<adios.file named {self._io_name}>"

    @singledispatchmethod
    def __init__(self, path, comm=None):
        super().__init__(path, "rra", comm)

    # e.g. FileReader(io: adios2.IO, path, mode)
    # pylint: disable=E1121
    @__init__.register(IO)
    def _(self, io: IO, path, comm=None):
        super().__init__(io, path, "rra", comm)

    # pylint: enable=E1121



from mpi4py import MPI
import numpy as np
from adios2 import Stream

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

nx = 10
shape = [size * nx]
start = [rank * nx]
count = [nx]

temperature = np.zeros(nx, dtype=np.double)
pressure = np.ones(nx, dtype=np.double)
delta_time = 0.01
physical_time = 0.0
nsteps = 5

with Stream("cfd.bp", "w", comm) as s:
    # NSteps from application
    for _ in s.steps(nsteps):
        if rank == 0 and s.current_step() == 0:
            # write a Python integer
            s.write("nproc", size)

        # write a Python floating point value
        s.write("physical_time", physical_time)
        # temperature and pressure are numpy arrays
        s.write("temperature", temperature, shape, start, count)
        s.write_attribute("temperature/unit", "K")
        s.write("pressure", pressure, shape, start, count)
        s.write_attribute("pressure/unit", "Pa")
        physical_time += delta_time



#!/usr/bin/env python
#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestBPWriteReadString.py: test writing/reading Python string type
# in ADIOS2 File Write
#  Created on: Oct 19, 2020
#      Author: Dmitry Ganyushin ganyushindi@ornl.gov
import unittest
from mpi4py import MPI
import adios2.bindings as adios2

N_STEPS = 3


class TestAdiosWriteReadStringfullAPI(unittest.TestCase):
    def test_write_read_string_fullAPI(self):
        comm = MPI.COMM_WORLD
        theString = "hello adios"
        bpFilename = "string_test_fullAPI.bp"
        varname = "mystringvar"
        adios = adios2.ADIOS(comm)
        ioWrite = adios.DeclareIO("ioWriter")
        adEngine = ioWrite.Open(bpFilename, adios2.Mode.Write)
        varMyString = ioWrite.DefineVariable(varname)
        for step in range(N_STEPS):
            adEngine.BeginStep()
            adEngine.Put(varMyString, theString + str(step))
            adEngine.EndStep()
        adEngine.Close()

        ioRead = adios.DeclareIO("ioReader")
        adEngine = ioRead.Open(bpFilename, adios2.Mode.Read)
        for step in range(N_STEPS):
            adEngine.BeginStep()
            varReadMyString = ioRead.InquireVariable(varname)
            result = adEngine.Get(varReadMyString)
            adEngine.EndStep()
            self.assertEqual(result, theString + str(step))
        adEngine.Close()


if __name__ == "__main__":
    unittest.main()



"""License:
  Distributed under the OSI-approved Apache License, Version 2.0.  See
  accompanying file Copyright.txt for details.
"""


class Attribute:
    """High level representation of the Attribute class in the adios2.bindings"""

    def __init__(self, io, name, *args, **kwargs):
        self.impl = io.DefineAttribute(name, *args, **kwargs)

    @property
    def impl(self):
        """Bindings implementation of the class"""
        return self._impl

    @impl.setter
    def impl(self, implementation):
        self._impl = implementation

    def __eq__(self, other):
        if isinstance(other, Attribute):
            return self.name() == other.name()
        return False

    def name(self):
        """
        Name of the Attribute

        Returns:
            Name of the Attribute as a str.
        """
        return self.impl.Name()

    def type(self):
        """
        Type of the Attribute

        Returns:
            Type of the Attribute as a str.
        """
        return self.impl.Type()

    def single_value(self):
        """
        True if the attribute is a single value, False if it is an array

        Returns:
            True or False.
        """
        return self.impl.SingleValue()

    def data(self):
        """
        Content of the Attribute

        Returns:
            Content of the Attribute as a non string.
        """
        if self.single_value():
            return self.impl.Data()[0]
        return self.impl.Data()

    def data_string(self):
        """
        Content of the Attribute

        Returns:
            Content of the Attribute as a str.
        """
        if self.single_value():
            return self.impl.DataString()[0]
        return self.impl.DataString()



#!/usr/bin/env python

#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestBPWriteTypes.py: test Python numpy types in ADIOS2 File
#                      Write/Read High-Level API
#  Created on: March 12, 2018
#      Author: William F Godoy godoywf@ornl.gov

from adios2NPTypes import SmallTestData
from mpi4py import MPI
import numpy as np
from adios2 import Stream, LocalValueDim

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# Test data
data = SmallTestData(rank)
nx = data.Nx

shape = [size * nx]
start = [rank * nx]
count = [nx]

# Writer
with Stream("types_np.bp", "w", comm=comm) as s:
    for step in s.steps(5):
        data.update(rank, step.current_step(), size)
        s.write("rank", np.array(rank), shape=[LocalValueDim])
        if rank == 0 and step.current_step() == 0:
            s.write("tag", "Testing ADIOS2 high-level API")
            s.write("nx", data.Nx)
            s.write("gvarI8", np.array(data.i8[0]))
            s.write("gvarI16", np.array(data.i16[0]))
            s.write("gvarI32", np.array(data.i32[0]))
            s.write("gvarI64", np.array(data.i64[0]))
            s.write("gvarU8", np.array(data.u8[0]))
            s.write("gvarU16", np.array(data.u16[0]))
            s.write("gvarU32", np.array(data.u32[0]))
            s.write("gvarU64", np.array(data.u64[0]))
            s.write("gvarR32", np.array(data.r32[0]))
            s.write("gvarR64", np.array(data.r64[0]))
            s.write("gvarC64", np.array(data.c64[0]))
            i = data.int_list[0]
            f = data.float_list[0]
            c = data.complex_list[0]
            print(f"type of i = {type(i)}, type of f = {type(f)}, type of c = {type(c)}")
            s.write("an_int_value", i)
            s.write("a_float_value", f)
            s.write("a_complex_value", c)
            s.write("an_int_list", data.int_list)
            s.write("a_float_list", data.float_list)
            s.write("a_complex_list", data.complex_list)

            # single value attributes with numpy variables
            s.write_attribute("attrStr", "Testing single string attribute")
            print(
                f"---- type of np.array(data.i8[0]) is {type(np.array(data.i8[0]))}"
                f" shape = {np.array(data.i8[0]).shape}"
            )
            s.write_attribute("attrI8", np.array(data.i8[0]))
            s.write_attribute("attrI16", np.array(data.i16[0]))
            s.write_attribute("attrI32", np.array(data.i32[0]))
            s.write_attribute("attrI64", np.array(data.i64[0]))
            s.write_attribute("attrU8", np.array(data.u8[0]))
            s.write_attribute("attrU16", np.array(data.u16[0]))
            s.write_attribute("attrU32", np.array(data.u32[0]))
            s.write_attribute("attrU64", np.array(data.u64[0]))
            s.write_attribute("attrR32", np.array(data.r32[0]))
            s.write_attribute("attrR64", np.array(data.r64[0]))
            s.write_attribute("attrC64", np.array(data.c64[0]))

            # single value attributes with Python variables
            s.write_attribute("attrNx", data.Nx)
            s.write_attribute("attr_int_value", i)
            s.write_attribute("attr_float_value", f)
            s.write_attribute("attr_complex_value", c)

            # array attributes with Python lists
            s.write_attribute("attr_int_list", data.int_list)
            s.write_attribute("attr_float_list", data.float_list)
            s.write_attribute("attr_complex_list", data.complex_list)
            s.write_attribute("attrStrArray1", ["string1"])
            s.write_attribute("attrStrArray3", ["string1", "string2", "string3"])

            # array attributes with numpy arrays
            s.write_attribute("attrI8Array", data.i8)
            s.write_attribute("attrI16Array", data.i16)
            s.write_attribute("attrI32Array", data.i32)
            s.write_attribute("attrI64Array", data.i64)
            s.write_attribute("attrU8Array", data.u8)
            s.write_attribute("attrU16Array", data.u16)
            s.write_attribute("attrU32Array", data.u32)
            s.write_attribute("attrU64Array", data.u64)
            s.write_attribute("attrR32Array", data.r32)
            s.write_attribute("attrR64Array", data.r64)
            s.write_attribute("attrC64Array", data.c64)

        s.write("steps", "Step:" + str(step.current_step()))
        s.write("varI8", data.i8, shape, start, count)
        s.write("varI16", data.i16, shape, start, count)
        s.write("varI32", data.i32, shape, start, count)
        s.write("varI64", data.i64, shape, start, count)
        s.write("varU8", data.u8, shape, start, count)
        s.write("varU16", data.u16, shape, start, count)
        s.write("varU32", data.u32, shape, start, count)
        s.write("varU64", data.u64, shape, start, count)
        s.write("varR32", data.r32, shape, start, count)
        s.write("varR64", data.r64, shape, start, count)
        s.write("varC64", data.c64, shape, start, count)

        if rank == 0 and step.current_step() == 0:
            # attribute assigned to variable
            s.write_attribute("size", data.Nx, "varI8")
            s.write_attribute("size", data.Nx, "varI16", "::")
            s.write_attribute("varattrStrArray", ["varattr1", "varattr2", "varattr3"], "steps")
            s.write_attribute("varattrI8Array", data.i8, "varI8")
            s.write_attribute("varattrI16Array", data.i16, "varI16")
            s.write_attribute("varattrI32Array", data.i32, "varI32")
            s.write_attribute("varattrI64Array", data.i64, "varI64")
            s.write_attribute("varattrU8Array", data.u8, "varU8")
            s.write_attribute("varattrU16Array", data.u16, "varU16")
            s.write_attribute("varattrU32Array", data.u32, "varU32")
            s.write_attribute("varattrU64Array", data.u64, "varU64")
            s.write_attribute("varattrR32Array", data.r32, "varR32")
            s.write_attribute("varattrR64Array", data.r64, "varR64")
            s.write_attribute("varattrR64Value", data.r64, "varR64")

comm.Barrier()

# Reader
data = SmallTestData(rank)

with Stream("types_np.bp", "r", comm=comm) as fr:
    # file only
    assert fr.num_steps() == 5

    for fr_step in fr.steps():
        step = fr_step.current_step()
        data.update(rank, step, size)

        step_vars = fr_step.available_variables()

        #         for name, info in step_vars.items():
        #             print("variable_name: " + name)
        #             for key, value in info.items():
        #                 print("\t" + key + ": " + value)
        #             print("\n")

        if rank == 0 and step == 0:
            inTag = fr_step.read("tag")
            print(f"tag = {inTag}")
            nx = fr_step.read("nx")
            print(f"nx = {nx}")
            assert nx == data.Nx
            inI8 = fr_step.read("gvarI8")
            inI16 = fr_step.read("gvarI16")
            inI32 = fr_step.read("gvarI32")
            inI64 = fr_step.read("gvarI64")
            inU8 = fr_step.read("gvarU8")
            inU16 = fr_step.read("gvarU16")
            inU32 = fr_step.read("gvarU32")
            inU64 = fr_step.read("gvarU64")
            inR32 = fr_step.read("gvarR32")
            inR64 = fr_step.read("gvarR64")

            if inTag != "Testing ADIOS2 high-level API":
                print("InTag: " + str(inTag))
                raise ValueError("tag variable read failed")

            if inI8 != data.i8[0]:
                raise ValueError("gvarI8 read failed")

            if inI16 != data.i16[0]:
                raise ValueError("gvarI16 read failed")

            if inI32 != data.i32[0]:
                raise ValueError("gvarI32 read failed")

            if inI64 != data.i64[0]:
                raise ValueError("gvarI64 read failed")

            if inU8 != data.u8[0]:
                raise ValueError("gvarU8 read failed")

            if inU16 != data.u16[0]:
                raise ValueError("gvarU16 read failed")

            if inU32 != data.u32[0]:
                raise ValueError("gvarU32 read failed")

            if inU64 != data.u64[0]:
                raise ValueError("gvarU64 read failed")

            if inR32 != data.r32[0]:
                raise ValueError("gvarR32 read failed")

            if inR64 != data.r64[0]:
                raise ValueError("gvarR64 read failed")

            # attributes
            inTag = fr_step.read_attribute("attrStr")
            print(f"attrStr = {inTag}")
            inNx = fr_step.read_attribute("attrNx")
            inI8 = fr_step.read_attribute("attrI8")
            inI16 = fr_step.read_attribute("attrI16")
            print(f"attrI16 = {inI16}")
            inI32 = fr_step.read_attribute("attrI32")
            inI64 = fr_step.read_attribute("attrI64")
            inU8 = fr_step.read_attribute("attrU8")
            inU16 = fr_step.read_attribute("attrU16")
            inU32 = fr_step.read_attribute("attrU32")
            inU64 = fr_step.read_attribute("attrU64")
            inR32 = fr_step.read_attribute("attrR32")
            inR64 = fr_step.read_attribute("attrR64")

            if inTag != "Testing single string attribute":
                raise ValueError("attr string read failed")

            if inNx != data.Nx:
                raise ValueError("attrNx read failed")

            if inI8 != data.i8[0]:
                raise ValueError("attrI8 read failed")

            if inI16 != data.i16[0]:
                raise ValueError("attrI16 read failed")

            if inI32 != data.i32[0]:
                raise ValueError("attrI32 read failed")

            if inI64 != data.i64[0]:
                raise ValueError("attrI64 read failed")

            if inU8 != data.u8[0]:
                raise ValueError("attrU8 read failed")

            if inU16 != data.u16[0]:
                raise ValueError("attrU16 read failed")

            if inU32 != data.u32[0]:
                raise ValueError("attrU32 read failed")

            if inU64 != data.u64[0]:
                raise ValueError("attrU64 read failed")

            if inR32 != data.r32[0]:
                raise ValueError("attrR32 read failed")

            if inR64 != data.r64[0]:
                raise ValueError("attrR64 read failed")

            in_an_int_value = fr_step.read("an_int_value")
            print(f"an_int_value = {in_an_int_value} of type {type(in_an_int_value)}")
            assert in_an_int_value == data.int_list[0]

            in_a_float_value = fr_step.read("a_float_value")
            print(f"a_float_value = {in_a_float_value} of type {type(in_a_float_value)}")
            assert in_a_float_value == data.float_list[0]

            inC64 = fr_step.read("gvarC64")
            print(f"inC64 = {inC64} of type {type(inC64)}")
            assert inC64 == data.c64[0]

            in_a_complex_value = fr_step.read("a_complex_value")
            print(f"a_complex_value = {in_a_complex_value} of type {type(in_a_complex_value)}")
            assert in_a_complex_value == data.complex_list[0]

            an_int_list = fr_step.read("an_int_list")
            if not (an_int_list == data.int_list).all():
                raise ValueError("an_int_list array read failed")
            print(f"an_int_list = {an_int_list} of type {type(an_int_list)}")

            a_float_list = fr_step.read("a_float_list")
            if not (a_float_list == data.float_list).all():
                raise ValueError("a_float_list array read failed")
            print(f"a_float_list = {a_float_list} of type {type(a_float_list)}")

            a_complex_list = fr_step.read("a_complex_list")
            if not (a_complex_list == data.complex_list).all():
                raise ValueError("a_complex_list array read failed")
            print(f"a_complex_list = {a_complex_list} of type {type(a_complex_list)}")

            # Array attribute
            inStr1 = fr_step.read_attribute("attrStrArray1")
            print(f"attrStrArray1 = {inStr1}")
            inStr3 = fr_step.read_attribute("attrStrArray3")
            print(f"attrStrArray3 = {inStr3}")
            inI8 = fr_step.read_attribute("attrI8Array")
            inI16 = fr_step.read_attribute("attrI16Array")
            print(f"attrI16Array = {inI16}")
            inI32 = fr_step.read_attribute("attrI32Array")
            inI64 = fr_step.read_attribute("attrI64Array")
            inU8 = fr_step.read_attribute("attrU8Array")
            inU16 = fr_step.read_attribute("attrU16Array")
            inU32 = fr_step.read_attribute("attrU32Array")
            inU64 = fr_step.read_attribute("attrU64Array")
            inR32 = fr_step.read_attribute("attrR32Array")
            inR64 = fr_step.read_attribute("attrR64Array")

            if inStr1 != ["string1"]:
                raise ValueError("attrStrArray1 read failed")

            if inStr3 != ["string1", "string2", "string3"]:
                raise ValueError("attrStrArray3 read failed")

            if not (inI8 == data.i8).all():
                raise ValueError("attrI8 array read failed")

            if not (inI16 == data.i16).all():
                raise ValueError("attrI16 array read failed")

            if not (inI32 == data.i32).all():
                raise ValueError("attrI32 array read failed")

            if not (inI64 == data.i64).all():
                raise ValueError("attrI64 array read failed")

            if not (inU8 == data.u8).all():
                raise ValueError("attrU8 array read failed")

            if not (inU16 == data.u16).all():
                raise ValueError("attrU16 array read failed")

            if not (inU32 == data.u32).all():
                raise ValueError("attrU32 array read failed")

            if not (inU64 == data.u64).all():
                raise ValueError("attrU64 array read failed")

            if not (inR32 == data.r32).all():
                raise ValueError("attrR32 array read failed")

            if not (inR64 == data.r64).all():
                raise ValueError("attrR64 array read failed")

            # Array attributes (written as List)
            in_attr_float_list = fr_step.read_attribute("attr_float_list")
            in_attr_int_list = fr_step.read_attribute("attr_int_list")

            print(f"attr_float_list = {in_attr_float_list}")
            print(f"attr_int_list = {in_attr_int_list}")

            if not (in_attr_float_list == np.array(data.float_list)).all():
                raise ValueError("attr_float_list array read failed")

            if not (in_attr_int_list == np.array(data.int_list)).all():
                raise ValueError("attr_int_list array read failed")

            inTags = fr_step.read_attribute_string("varattrStrArray", "steps")
            inI8 = fr_step.read_attribute("varattrI8Array", "varI8")
            in16 = fr_step.read_attribute("varattrI16Array", "varI16")
            inI32 = fr_step.read_attribute("varattrI32Array", "varI32")
            inI64 = fr_step.read_attribute("varattrI64Array", "varI64")
            inU8 = fr_step.read_attribute("varattrU8Array", "varU8")
            inU16 = fr_step.read_attribute("varattrU16Array", "varU16")
            inU32 = fr_step.read_attribute("varattrU32Array", "varU32")
            inU64 = fr_step.read_attribute("varattrU64Array", "varU64")
            inR32 = fr_step.read_attribute("varattrR32Array", "varR32")
            inR64 = fr_step.read_attribute("varattrR64Array", "varR64")

            if inTags != ["varattr1", "varattr2", "varattr3"]:
                print(inTags)
                raise ValueError("var attrStrArray read failed")

            if not (inI8 == data.i8).all():
                raise ValueError("var attrI8 array read failed")

            if not (inI16 == data.i16).all():
                raise ValueError("var attrI16 array read failed")

            if not (inI32 == data.i32).all():
                raise ValueError("var attrI32 array read failed")

            if not (inI64 == data.i64).all():
                raise ValueError("var attrI64 array read failed")

            if not (inU8 == data.u8).all():
                raise ValueError("var attrU8 array read failed")

            if not (inU16 == data.u16).all():
                raise ValueError("var attrU16 array read failed")

            if not (inU32 == data.u32).all():
                raise ValueError("var attrU32 array read failed")

            if not (inU64 == data.u64).all():
                raise ValueError("var attrU64 array read failed")

            if not (inR32 == data.r32).all():
                raise ValueError("var attrR32 array read failed")

            if not (inR64 == data.r64).all():
                raise ValueError("var attrR64 array read failed")

            # Attributes assigned to a variable
            sizeI8 = fr_step.read_attribute("size", "varI8")
            sizeI16 = fr_step.read_attribute("size", "varI16", "::")

            if sizeI8 != data.Nx:
                raise ValueError("attribute varI8/size read failed")

            if sizeI16 != data.Nx:
                raise ValueError("attribute varI16::size read failed")

            sizeI8 = fr_step.read_attribute("varI8/size")
            sizeI16 = fr_step.read_attribute("varI16::size")

            if sizeI8 != data.Nx:
                raise ValueError("attribute varI8/size read failed")

            if sizeI16 != data.Nx:
                raise ValueError("attribute varI16::size read failed")

            step_attrs = fr_step.available_attributes()
            # for name, info in step_attrs.items():
            #     print(f"attribute {name} : {info}")

            step_attrs = fr_step.available_attributes("varI8")
            # for name, info in step_attrs.items():
            #    print(f"attribute {name} : {info}")
            if not [*step_attrs] == ["size", "varattrI8Array"]:
                raise ValueError("getting attributes of varI8 failed")

            step_attrs = fr_step.available_attributes("varI16", "::")
            for name, info in step_attrs.items():
                print(f"attribute {name} : {info}")
            if not [*step_attrs] == ["size"]:
                raise ValueError("getting attributes of varI16 with separator :: failed")

        stepStr = "Step:" + str(step)

        instepStr = fr_step.read("steps")
        if instepStr != stepStr:
            raise ValueError("steps variable read failed: " + instepStr + " " + stepStr)

        indataRanks = fr_step.read("rank", [0], [size])
        dataRanks = np.arange(0, size)
        if not (indataRanks == dataRanks).all():
            raise ValueError("Ranks read failed")

        indataI8 = fr_step.read("varI8", start, count)
        indataI16 = fr_step.read("varI16", start, count)
        indataI32 = fr_step.read("varI32", start, count)
        indataI64 = fr_step.read("varI64", start, count)
        indataU8 = fr_step.read("varU8", start, count)
        indataU16 = fr_step.read("varU16", start, count)
        indataU32 = fr_step.read("varU32", start, count)
        indataU64 = fr_step.read("varU64", start, count)
        indataR32 = fr_step.read("varR32", start, count)
        indataR64 = fr_step.read("varR64", start, count)

        if not (indataI8 == data.i8).all():
            raise ValueError("i8 array read failed")

        if not (indataI16 == data.i16).all():
            raise ValueError("i16 array read failed")

        if not (indataI32 == data.i32).all():
            raise ValueError("i32 array read failed")

        if not (indataI64 == data.i64).all():
            raise ValueError("i64 array read failed")

        if not (indataU8 == data.u8).all():
            raise ValueError("u8 array read failed")

        if not (indataU16 == data.u16).all():
            raise ValueError("u16 array read failed")

        if not (indataU32 == data.u32).all():
            raise ValueError("u32 array read failed")

        if not (indataU64 == data.u64).all():
            raise ValueError("u64 array read failed")

        if not (indataR32 == data.r32).all():
            raise ValueError("r32 array read failed")

        if not (indataR64 == data.r64).all():
            raise ValueError("r64 array read failed")



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestNullEngine.py
#
#
# Created on : Apr 11th, 2019
# Author : William F Godoy godoywf @ornl.gov
#

from mpi4py import MPI
import numpy as np
import adios2.bindings as adios2

# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# User data
Nx = 10
Ny = 10

count = [Nx, Ny]
start = [rank * Nx, 0]
shape = [size * Nx, Ny]

temperatures = np.zeros(count, dtype=np.int32)

for i in range(0, Nx):
    for j in range(0, Ny):
        temperatures[i, j] = (start[0] + i) * shape[1] + (j + start[1])

# ADIOS write
adios = adios2.ADIOS(comm)
ioWrite = adios.DeclareIO("ioWriter")

varTemperature = ioWrite.DefineVariable(
    "temperature2D", temperatures, shape, start, count, adios2.ConstantDims
)
ioWrite.SetEngine("NULL")

nullWriter = ioWrite.Open("NULL_py.bp", adios2.Mode.Write)

assert nullWriter.Type() == "NullWriter"

status = nullWriter.BeginStep()
assert status == adios2.StepStatus.OK

nullWriter.Put(varTemperature, temperatures)
nullWriter.EndStep()
nullWriter.Close()

# ADIOS2 read
ioRead = adios.DeclareIO("ioReader")
ioRead.SetEngine("null")
nullReader = ioRead.Open("NULL_py.bp", adios2.Mode.Read, MPI.COMM_SELF)

assert nullReader.Type() == "NullReader"

inTemperatures = np.zeros(1, dtype=np.int32)

status = nullReader.BeginStep()
assert status == adios2.StepStatus.EndOfStream

var_inTemperature = ioRead.InquireVariable("temperature2D")

if var_inTemperature is True:
    raise ValueError("var_inTemperature is not False")

# nullReader.Get(var_inTemperature, inTemperatures)

nullReader.PerformGets()
nullReader.EndStep()
nullReader.Close()



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# Test writer to produce a test dataset for Matlab reading
# This is serial since the Matlab reader is built with serial ADIOS
#
#
#  Created on: Jan 28, 2019
#      Author: Norbert Podhorszki, pnorbert@ornl.gov
#

import numpy
import adios2

# User data
NRows = 5
NCols = 6

shape = [NRows, NCols]
start = [0, 0]
count = [NRows, NCols]

temperatures = numpy.zeros(NRows * NCols, dtype=numpy.int16)

value = (NRows * NCols) + 1
for i in range(0, NRows):
    for j in range(0, NCols):
        temperatures[i * NCols + j] = value
        value = value + 1

# ADIOS2 high-level API for Write
fw = adios2.open("test1.bp", "w")
fw.write("note", 'This is an ADIOS2 output')
fw.write("temperature2D", temperatures, shape, start, count)
fw.write("nrows", numpy.array([NRows]))
fw.write("ncols", numpy.array([NCols]))
fw.close()



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# bpReaderHeatMap2D-bindings.py
#
#
#  Created on: Dec 5th, 2017
#      Author: William F Godoy godoywf@ornl.gov
#              Norbert Podhorszki pnorbert@ornl.gov
#

from mpi4py import MPI
import numpy
import adios2.bindings as adios2

# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# User data
Nx = 10
Ny = 10

count = [Nx, Ny]
start = [rank * Nx, 0]
shape = [size * Nx, Ny]

temperatures = numpy.zeros(Nx * Ny, dtype=int)

for i in range(0, Nx):
    iGlobal = start[0] + i

    for j in range(0, Ny):
        value = iGlobal * shape[1] + j
        temperatures[i * Nx + j] = value
#        print(str(i) + "," + str(j) + " " + str(value))


# ADIOS portion
adios = adios2.ADIOS(comm)
ioWrite = adios.DeclareIO("ioWriter")

varTemperature = ioWrite.DefineVariable(
    "temperature2D", temperatures, shape, start, count, adios2.ConstantDims
)

obpStream = ioWrite.Open("HeatMap2D_py_bindings.bp", adios2.Mode.Write)
obpStream.BeginStep()
obpStream.Put(varTemperature, temperatures)
obpStream.EndStep()
obpStream.Close()


if rank == 0:
    ioRead = adios.DeclareIO("ioReader")

    ibpStream = ioRead.Open("HeatMap2D_py_bindings.bp", adios2.Mode.Read, MPI.COMM_SELF)

    ibpStream.BeginStep()

    var_inTemperature = ioRead.InquireVariable("temperature2D")

    if var_inTemperature is not None:
        var_inTemperature.SetSelection([[2, 2], [4, 4]])

        inSize = var_inTemperature.SelectionSize()
        print("Incoming size " + str(inSize))
        inTemperatures = numpy.zeros(var_inTemperature.Count(), dtype=int)

        ibpStream.Get(var_inTemperature, inTemperatures, adios2.Mode.Sync)

        print("Incoming temperature map")
        for i in range(0, inTemperatures.shape[1]):
            print(str(inTemperatures[i]) + " ")

    ibpStream.EndStep()
    ibpStream.Close()



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# Test writer to produce a test dataset for Matlab reading
# This is serial since the Matlab reader is built with serial ADIOS
#
#
#  Created on: Jan 28, 2019
#      Author: Norbert Podhorszki, pnorbert@ornl.gov
#

import numpy
import adios2

# User data
NRows = 5
NCols = 6

shape = [NRows, NCols]
start = [0, 0]
count = [NRows, NCols]

temperatures = numpy.zeros(NRows * NCols, dtype=numpy.int16)
npcols = numpy.array([NCols])
nprows = numpy.array([NRows])

value = (NRows * NCols) + 1
for i in range(0, NRows):
    for j in range(0, NCols):
        temperatures[i * NCols + j] = value
        value = value + 1

# ADIOS2 low-level API for Write
adios = adios2.ADIOS()
io = adios.DeclareIO("writer")

varStr = io.DefineVariable("note")
varCols = io.DefineVariable("ncols", npcols)
varRows = io.DefineVariable("nrows", nprows)
varT = io.DefineVariable(
    "temperature2D", temperatures, shape, start, count, adios2.ConstantDims)

io.DefineAttribute("aaa", numpy.array([3.1415]))
io.DefineAttribute("anote", "just a string")
io.DefineAttribute("adimnames", ["rows", "columns"])
npdims = numpy.array([NRows, NCols], dtype=numpy.int32)
io.DefineAttribute("adims", npdims)


fw = io.Open("test1.bp", adios2.Mode.Write)
fw.Put(varStr, "This is an ADIOS2 output")
fw.Put(varCols, npcols)
fw.Put(varRows, nprows)
fw.Put(varT, temperatures)
fw.Close()



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# hello-world.py : adios2 low-level API example to write and read a
#                   string Variable with a greeting
#
#  Created on: 2/2/2021
#      Author: Dmitry Ganyushin ganyushindi@ornl.gov
#
import sys
from mpi4py import MPI
import adios2.bindings as adios2

DATA_FILENAME = "hello-world-py-bindings.bp"
# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()


def writer(ad, greeting):
    """write a string to a bp file"""
    io = ad.DeclareIO("hello-world-writer")
    var_greeting = io.DefineVariable("Greeting")
    w = io.Open(DATA_FILENAME, adios2.Mode.Write)
    w.BeginStep()
    w.Put(var_greeting, greeting)
    w.EndStep()
    w.Close()
    return 0


def reader(ad):
    """read a string from to a bp file"""
    io = ad.DeclareIO("hello-world-reader")
    r = io.Open(DATA_FILENAME, adios2.Mode.Read)
    r.BeginStep()
    var_greeting = io.InquireVariable("Greeting")
    message = r.Get(var_greeting)
    r.EndStep()
    r.Close()
    return message


def main():
    """driver function"""
    ad = adios2.ADIOS(comm)
    greeting = "Hello World from ADIOS2"
    writer(ad, greeting)
    message = reader(ad)
    print("{}".format(message))
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestBPWriteTypes.py: test Python numpy types in ADIOS2 File Write
#  Created on: Feb 2, 2017
#      Author: William F Godoy godoywf@ornl.gov


from adios2NPTypes import SmallTestData
import adios2.bindings as adios2


# Test data
data = SmallTestData()

adios = adios2.ADIOS()

bpIO = adios.DeclareIO("NPTypes")

# ADIOS Variable name, shape, start, offset, constant dims
# All local variables
varI8 = bpIO.DefineVariable("varI8", data.I8, [], [], [data.I8.size], adios2.ConstantDims)
varI16 = bpIO.DefineVariable("varI16", data.I16, [], [], [data.I16.size], adios2.ConstantDims)
varI32 = bpIO.DefineVariable("varI32", data.I32, [], [], [data.I32.size], adios2.ConstantDims)
varI64 = bpIO.DefineVariable("varI64", data.I64, [], [], [data.I64.size], adios2.ConstantDims)

varU8 = bpIO.DefineVariable("varUI8", data.U8, [], [], [data.U8.size], adios2.ConstantDims)
varU16 = bpIO.DefineVariable("varUI16", data.U16, [], [], [data.U16.size], adios2.ConstantDims)
varU32 = bpIO.DefineVariable("varUI32", data.U32, [], [], [data.U32.size], adios2.ConstantDims)
varU64 = bpIO.DefineVariable("varUI64", data.U64, [], [], [data.U64.size], adios2.ConstantDims)

varR32 = bpIO.DefineVariable("varR32", data.R32, [], [], [data.R32.size], adios2.ConstantDims)

varR64 = bpIO.DefineVariable("varR64", data.R64, [], [], [data.R64.size], adios2.ConstantDims)


# ADIOS Engine
bpFileWriter = bpIO.Open("npTypes.bp", adios2.Mode.Write)

bpFileWriter.Put(varI8, data.I8)
bpFileWriter.Put(varI16, data.I16)
bpFileWriter.Put(varI32, data.I32)
bpFileWriter.Put(varI64, data.I64)

bpFileWriter.Put(varU8, data.U8)
bpFileWriter.Put(varU16, data.U16)
bpFileWriter.Put(varU32, data.U32)
bpFileWriter.Put(varU64, data.U64)

bpFileWriter.Put(varR32, data.R32)
bpFileWriter.Put(varR64, data.R64)

bpFileWriter.Close()



import sys
import json
import argparse


def searchForContext(context):
    print('Searching for a status with context: %s' % context)
    statuses = json.load(sys.stdin)

    for stat in statuses:
        if 'context' in stat and stat['context'] == context:
            sys.exit(0)

    sys.exit(1)


# =============================================================================
# Main
# =============================================================================

if __name__ == "__main__":
    parser = argparse.ArgumentParser("Parse github api status list")
    parser.add_argument("--context", default=None, help="context of interest")
    args = parser.parse_args()
    searchForContext(args.context)



#!/usr/bin/env python

#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestBPPNGHighLevelAPI.py
#
#  Created on: June 7th, 2019
#      Author: William F Godoy

import numpy as np
import random
from mpi4py import MPI
from adios2 import Stream


def compress_png(compression_level):
    fname = "BPWRPNG_" + str(compression_level) + "_py.bp"
    Nx = 10
    Ny = 50
    channels = 3
    NSteps = 1

    # initialize values
    u32s = np.zeros([Nx, Ny], np.uint32)
    u8s = np.zeros([Nx, Ny, channels], np.uint8)

    value_ji = 0.0
    for i in range(0, Nx):
        for j in range(0, Ny):
            u32s[i][j] = value_ji
            u8s[i][j][0] = random.randrange(256)
            u8s[i][j][1] = random.randrange(256)
            u8s[i][j][2] = random.randrange(256)

            value_ji += 1.0

    # set global dimensions
    # MPI
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    shape3D = [Nx * size, Ny, 3]
    start3D = [Nx * rank, 0, 0]
    count3D = [Nx, Ny, 3]

    shape2D = [Nx * size, Ny]
    start2D = [Nx * rank, 0]
    count2D = [Nx, Ny]

    # writer
    with Stream(fname, "w", comm=comm) as s:
        for step in s.steps(NSteps):
            s.write(
                "u8",
                u8s,
                shape3D,
                start3D,
                count3D,
                [
                    (
                        "PNG",
                        {
                            "bit_depth": "8",
                            "color_type": "PNG_COLOR_TYPE_RGB",
                            "compression_level": str(compression_level),
                        },
                    )
                ],
            )
            s.write(
                "u32",
                u32s,
                shape2D,
                start2D,
                count2D,
                [
                    (
                        "PNG",
                        {
                            "bit_depth": "8",
                            "color_type": "PNG_COLOR_TYPE_RGBA",
                            "compression_level": str(compression_level),
                        },
                    )
                ],
            )

    # reader
    with Stream(fname, "r", comm=comm) as s:
        for step in s.steps():
            in_u8s = step.read("u8", start3D, count3D)
            in_u32s = step.read("u32", start2D, count2D)

            for i in range(0, Nx):
                for j in range(0, Ny):
                    assert u32s[i][j] == in_u32s[i][j]
                    assert u8s[i][j][0] == in_u8s[i][j][0]
                    assert u8s[i][j][1] == in_u8s[i][j][1]
                    assert u8s[i][j][2] == in_u8s[i][j][2]


def main():
    compress_png(compression_level=1)
    compress_png(compression_level=4)
    compress_png(compression_level=9)


if __name__ == "__main__":
    main()



from adios2.adios import Adios
import adios2.bindings as bindings

import unittest


class TestIO(unittest.TestCase):
    def test_io_empty(self):
        adios = Adios()
        adios.declare_io("BPWriter")

    def test_io_define_attribute(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        ts = io.define_attribute("timestamp", "20231122")
        self.assertIsNot(ts, None)

    def test_io_inquire_attribute(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        ts = io.define_attribute("timestamp", "20231122")
        coords = io.define_attribute("coords", "43N74W")
        x = io.inquire_attribute("coords")
        self.assertNotEqual(ts, coords)
        self.assertNotEqual(ts, x)
        self.assertEqual(coords, x)
        self.assertIs(io.inquire_attribute("notanattribute"), None)

    def test_available_attributes(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        io.define_attribute("timestamp", "20231122")
        io.define_attribute("stringarray", ["one", "two", "three"])
        io.define_attribute("afloat", 3.14)
        io.define_attribute("floatarray", [3.14, 6.28])
        attrs = io.available_attributes()
        print("Available attributes:")
        for aname, ainfo in attrs.items():
            print(f"    {aname}:\t{ainfo}")
        self.assertEqual(attrs["timestamp"]["Value"], '"20231122"')
        self.assertEqual(attrs["stringarray"]["Value"], '{ "one", "two", "three" }')
        self.assertFalse("coords" in attrs)

    def test_remove_attribute(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        io.define_attribute("timestamp", "20231122")
        io.remove_attribute("timestamp")
        self.assertIs(io.inquire_attribute("timestamp"), None)

    def test_remove_all_attribute(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        io.define_attribute("timestamp", "20231122")
        io.define_attribute("coords", "43N74W")
        io.remove_all_attributes()
        self.assertIs(io.inquire_attribute("timestamp"), None)
        self.assertIs(io.inquire_attribute("coords"), None)

    def test_io_define_variable(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        temp = io.define_variable("temp")
        self.assertNotEqual(temp, None)

    def test_io_inquire_variable(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        temp = io.define_variable("temp")
        presure = io.define_variable("pressure")
        x = io.inquire_variable("pressure")
        self.assertNotEqual(temp, presure)
        self.assertNotEqual(temp, x)
        self.assertEqual(presure, x)

    def test_available_variable(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        io.define_variable("temp")
        io.inquire_variable("temp")
        self.assertIs(io.inquire_attribute("pressure"), None)

    def test_remove_variable(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        io.define_variable("temp")
        io.remove_variable("temp")
        self.assertIs(io.inquire_attribute("temp"), None)

    def test_remove_all_variable(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        io.define_variable("temp")
        io.define_variable("pressure")
        io.remove_all_variables()
        self.assertIs(io.inquire_attribute("pressure"), None)
        self.assertIs(io.inquire_attribute("temp"), None)

    def test_open_engine(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        io.set_engine("BPFile")
        io.set_parameter("threads", "2")
        io.set_parameters({"AsyncOpen": "On", "MaxOpenFilesAtOnce": "512"})
        io.add_transport("File", {"Library": "POSIX"})
        engine = io.open("pythontest.bp", bindings.Mode.Write)
        self.assertNotEqual(engine, None)
        engine.close()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestChangingShape.py
#
#  Created on: April 2nd, 2019
#      Author: Jeremy Logan

import numpy as np
from mpi4py import MPI
from adios2 import Stream

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# Test data
nx = [10, 15]
data = [np.zeros(nx[0]), np.ones(nx[1])]
shape = [[size * nx[0]], [size * nx[1]]]
start = [[rank * nx[0]], [rank * nx[1]]]
count = [[nx[0]], [nx[1]]]

# Write different sized arrays as separate steps
with Stream("out.bp", "w", comm=comm) as s:
    s.begin_step()
    s.write("z", data[0], shape[0], start[0], count[0])
    s.end_step()
    s.begin_step()
    s.write("z", data[1], shape[1], start[1], count[1])
    s.end_step()

# Read back arrays
with Stream("out.bp", "r", comm=comm) as s:
    for step in s.steps():
        shape_z = int(step.available_variables()["z"]["Shape"])
        print(shape_z)
        assert shape_z == int(shape[step.current_step()][0])



from adios2 import FileReader, Stream, LocalValueDim
from random import randint

import unittest


class TestFileReader(unittest.TestCase):
    def test_basic(self):
        with Stream("pythonfiletest.bp", "w") as s:
            for _ in s.steps(10):
                if s.current_step() == 0:
                    # String
                    s.write("Outlook", "Good")
                    # Global Value single step
                    s.write("LocationID", 42)
                # Global Array
                s.write(
                    "Temp",
                    content=[randint(15, 35), randint(15, 35), randint(15, 35)],
                    shape=[3],
                    start=[0],
                    count=[3],
                )
                # Local Value
                s.write("Wind", [5], shape=[LocalValueDim])
                # Local Array
                s.write("Coords", [38, -46], [], [], [2])
                # Global Value every step
                s.write("Hour", 8 + s.current_step())

        with FileReader("pythonfiletest.bp") as f:
            self.assertEqual(len(f.all_blocks_info("temp")), 10)
            outlook_var = f.inquire_variable("Outlook")
            outlook = f.read(outlook_var)
            self.assertEqual(outlook, "Good")

            for name in f.available_variables():
                var = f.inquire_variable(name)
                if not var.type() == "string":
                    if var.steps() > 1:
                        self.assertEqual(var.steps(), 10)
                    var.set_step_selection([0, var.steps()])
                    output = f.read(var)
                    print(f"var:{var.name()} output:{output}")

        with FileReader("pythonfiletest.bp") as f:
            output = f.read("LocationID")
            self.assertEqual(output.ndim, 0)
            self.assertEqual(output.size, 1)
            self.assertEqual(output, 42)

            output = f.read("Hour", step_selection=[0, 10])
            self.assertEqual(output.ndim, 1)
            self.assertEqual(output.size, 10)
            self.assertTrue((output == [8, 9, 10, 11, 12, 13, 14, 15, 16, 17]).all())

            output = f.read("Hour", step_selection=[5, 1])
            self.assertEqual(output.ndim, 1)
            self.assertEqual(output.size, 1)
            self.assertEqual(output, [13])

            output = f.read("Temp", step_selection=[0, 10])
            self.assertEqual(len(output), 30)
            print(f"var:Temp output:{output}")

            output = f.read("Temp", step_selection=[0, 5])
            self.assertEqual(len(output), 15)
            print(f"var:Temp output:{output}")

            output = f.read("Temp", start=[0], count=[2], step_selection=[0, 10])
            self.assertEqual(len(output), 20)
            print(f"var:Temp output:{output}")


if __name__ == "__main__":
    unittest.main()



from mpi4py import MPI
import numpy as np
from time import sleep
from adios2 import Stream, Adios, bindings

# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# ADIOS MPI Communicator
adios = Adios(comm)

# ADIOS IO
io = adios.declare_io("myIO")
io.set_engine("Sst")

# Python MPI is usually not compatible with MPI used to build ADIOS
# Must use TCP based data transport, or RDMA (ucx, libfabric) if available
io.set_parameter("DataTransport", "WAN")

# note: we need to use np.float32 to be compatible with data from C++ writer
# using "float" works in Python only but leads to type mismatch with C++
myArray = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0], dtype=np.float32)
myArray = 10.0 * rank + myArray
nx = len(myArray)
increment = nx * size * 1.0

with Stream(io, "helloSst", "w", comm) as stream:
    for _ in stream.steps(4):
        currentStep = stream.current_step()

        # imitating computation
        sleep(1.0)

        stream.write("bpFloats", myArray, [size * nx], [rank * nx], [nx])
        print("Rank=", rank, "loop index =", currentStep, "data =", myArray, flush=True)
        myArray += increment
        # Warning: the data of the current step is not published until
        # the next loop entry or the exit of the loop



#!/usr/bin/env python

#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestBPZfpHighLevelAPI.py
#
#  Created on: April 2nd, 2019
#      Author: William F Godoy

import numpy as np
from mpi4py import MPI
from adios2 import Stream


def CompressZfp2D(rate):
    fname = "BPWRZfp2D_" + str(rate) + "_py.bp"
    Nx = 100
    Ny = 50
    NSteps = 2

    # initialize values
    r32s = np.zeros([Ny, Nx], np.float32)
    r64s = np.zeros([Ny, Nx], np.float64)

    value_ji = 0.0
    for j in range(0, Ny):
        for i in range(0, Nx):
            r32s[j][i] = value_ji
            r64s[j][i] = value_ji
            value_ji += 1.0

    # set global dimensions
    # MPI
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    shape = [Ny * size, Nx]
    start = [Ny * rank, 0]
    count = [Ny, Nx]

    # writer
    with Stream(fname, "w", comm=comm) as s:
        for _ in s.steps(NSteps):
            s.write("r32", r32s, shape, start, count, [("zfp", {"accuracy": str(rate)})])
            s.write("r64", r64s, shape, start, count, [("zfp", {"accuracy": str(rate)})])

    # reader
    with Stream(fname, "r", comm=comm) as s:
        for _ in s.steps():
            in_r32s = s.read("r32", start, count)
            in_r64s = s.read("r64", start, count)

            for j in range(0, Ny):
                for i in range(0, Nx):
                    assert abs(r32s[j][i] - in_r32s[j][i]) < 1e-4
                    assert abs(r64s[j][i] - in_r64s[j][i]) < 1e-4


def main():
    CompressZfp2D(rate=8)
    CompressZfp2D(rate=9)
    CompressZfp2D(rate=10)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestBPWriteReadString.py: test writing/reading Python string type
# in ADIOS2 File Write
#  Created on: Oct 19, 2020
#      Author: Dmitry Ganyushin ganyushindi@ornl.gov
import unittest
from mpi4py import MPI
from adios2 import Stream

N_STEPS = 3


class test_adios_write_read_string_full_api(unittest.TestCase):
    def test_write_read_string_high_api(self):
        comm = MPI.COMM_WORLD
        theString = "hello adios"
        bpFilename = "string_test_highAPI.bp"
        varname = "mystringvar"

        with Stream(bpFilename, "w", comm=comm) as s:
            for step in s.steps(N_STEPS):
                s.write(varname, theString + str(step.current_step()))

        with Stream(bpFilename, "r", comm=comm) as s:
            for _ in s.steps():
                step = s.current_step()
                result = s.read(varname)
                self.assertEqual(result, theString + str(step))

    def test_read_strings_all_steps(self):
        comm = MPI.COMM_WORLD
        fileName = "string_test_all.bp"
        with Stream(fileName, "w", comm=comm) as s:
            i = 0
            for _ in s.steps(N_STEPS):
                s.write("string_variable", "written {}".format(i))
                i += 1

        # with Stream(fileName, "rra", comm = comm) as s:
        #    n = s.num_steps()
        #    name = "string_variable"
        #    result = s.read_string(name, 0, n)
        #    expected_str = ["written {}".format(i) for i in range(n)]
        #    self.assertEqual(result, expected_str)


if __name__ == "__main__":
    unittest.main()



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestBPWriteRead2D.py
#
#
#  Created on: Mar 3rd, 2019
#      Author: Kai Germaschewski <kai.germaschewski@unh.edu>
#              William F Godoy godoywf@ornl.gov
#

from mpi4py import MPI
import numpy as np
import adios2.bindings as adios2

# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# User data
Nx = 10
Ny = 10

count = [Nx, Ny]
start = [rank * Nx, 0]
shape = [size * Nx, Ny]

temperatures = np.zeros(count, dtype=np.int32)

for i in range(0, Nx):
    for j in range(0, Ny):
        temperatures[i, j] = (start[0] + i) * shape[1] + (j + start[1])

# print(temperatures)
# ADIOS2 read
adios = adios2.ADIOS(comm)
ioWrite = adios.DeclareIO("ioWriter")

varTemperature = ioWrite.DefineVariable(
    "temperature2D", temperatures, shape, start, count, adios2.ConstantDims
)

obpStream = ioWrite.Open("HeatMap2D_py.bp", adios2.Mode.Write)
obpStream.Put(varTemperature, temperatures)
obpStream.Close()


if rank == 0:
    # ADIOS2 read
    ioRead = adios.DeclareIO("ioReader")
    ibpStream = ioRead.Open("HeatMap2D_py.bp", adios2.Mode.ReadRandomAccess, MPI.COMM_SELF)
    var_inTemperature = ioRead.InquireVariable("temperature2D")

    if var_inTemperature is False:
        raise ValueError("var_inTemperature is False")

    assert var_inTemperature is not None
    readOffset = [2, 2]
    readSize = [4, 4]

    var_inTemperature.SetSelection([readOffset, readSize])
    inTemperatures = np.zeros(readSize, dtype=np.int32)
    ibpStream.Get(var_inTemperature, inTemperatures, adios2.Mode.Sync)
    ibpStream.Close()

    # print('Incoming temperature map\n', inTemperatures)
    expected = np.array(
        [[22, 23, 24, 25], [32, 33, 34, 35], [42, 43, 44, 45], [52, 53, 54, 55]], np.int32
    )
    assert np.array_equal(inTemperatures, expected)



#!/usr/bin/env python
#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestBPSelectSteps_nompi.py: test step selection by reading in Python
# in ADIOS2 File Write
#  Created on: Jan 29, 2021
#      Author: Dmitry Ganyushin ganyushindi@ornl.gov
import unittest
import numpy as np
from mpi4py import MPI
import adios2.bindings as adios2

TESTDATA_FILENAME = "steps_int32.bp"

# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()
Nx = 8
shape = [size * Nx]
start = [rank * Nx]
count = [Nx]


class TestAdiosSelectSteps(unittest.TestCase):
    def setUp(self):
        total_steps = 10
        with adios2.open(TESTDATA_FILENAME, "w", comm) as fh:
            for i in range(total_steps):
                fh.write("step", np.full((Nx), i, dtype=np.int32), shape, start, count)
                fh.end_step()

    def test_select_steps_reading_fullAPI(self):
        selected_steps = [3, 5, 7]
        param_string = ",".join([str(i) for i in selected_steps])
        adios = adios2.ADIOS()
        ioReadBP = adios.DeclareIO("hellopy")
        ioReadBP.SetParameter(TESTDATA_FILENAME, param_string)
        fh = ioReadBP.Open(TESTDATA_FILENAME, adios2.Mode.ReadRandomAccess, comm)
        var = ioReadBP.InquireVariable("step")
        var.SetSelection([[0], [size * Nx]])
        var.SetStepSelection([0, len(selected_steps)])
        data = np.zeros((len(selected_steps), size * Nx), dtype=np.int32)
        fh.Get(var, data)
        fh.PerformGets()
        self.assertTrue(
            all(
                [
                    list(data[i]) == [selected_steps[i] for x in range(len(data[i]))]
                    for i in range(len(selected_steps))
                ]
            )
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
import logging
import adios2.bindings as adios2

if __name__ == "__main__":
    __spec__ = None


def main():
    print("====================================")

    format = "%(asctime)s: %(message)s"
    logging.basicConfig(format=format, level=logging.INFO, datefmt="%H:%M:%S")
    writer()
    reader_fail()
    reader_success()


def writer():
    logging.info(" Writer: initiating writing")
    data = np.arange(3, dtype=np.float32)
    logging.info(f"        data dtypes: {data.dtype!s}")
    shape = data.shape
    count = shape
    start = (0,) * len(shape)
    logging.info(f"        data on writer side {data!s}")

    adios_io = adios2.ADIOS()
    IO = adios_io.DeclareIO("writer")

    writer = IO.Open("testdatafile", adios2.Mode.Write)
    writebuffer = IO.DefineVariable("np_data", data, shape, start, count, adios2.ConstantDims)
    if writebuffer:
        writer.BeginStep()
        writer.Put(writebuffer, data, adios2.Mode.Sync)
        writer.EndStep()
    else:
        raise ValueError("DefineVariable failed")

    writer.Close()

    logging.info(" Writer: writing finished")


def reader_fail():
    adios_io = adios2.ADIOS()
    io = adios_io.DeclareIO("reader")

    logging.info(" Reader: initiating erroneous reading ")
    reader = io.Open("testdatafile", adios2.Mode.Read)
    while True:
        stepStatus = reader.BeginStep()
        if stepStatus == adios2.StepStatus.OK:
            # inquire for variable
            recvar = io.InquireVariable("np_data")
            if recvar:
                got_exception = 0
                # determine the shape of the data that will be sent
                bufshape = recvar.Shape()
                typ = recvar.Type()
                logging.info("        Incoming variable type is " + typ)
                # allocate buffer for new numpy
                data = np.ones(bufshape)
                logging.info(f"        receive buffer  dtype: {data.dtype!s}")
                try:
                    reader.Get(recvar, data, adios2.Mode.Sync)
                except Exception as e:
                    got_exception = 1
                    logging.info(str(e))
                if not got_exception:
                    raise Exception("Didn't get an exception as expected")
                else:
                    logging.info("        Got expected exception")
            else:
                raise ValueError("InquireVariable failed")
        elif stepStatus == adios2.StepStatus.EndOfStream:
            break
        else:
            raise StopIteration(f"next step failed to initiate {stepStatus!s}")
        reader.EndStep()
    reader.Close()
    logging.info(
        " Reader: finished reading",
    )


def reader_success():
    adios_io = adios2.ADIOS()
    io = adios_io.DeclareIO("reader")

    logging.info(" Reader: initiating successful reading ")
    reader = io.Open("testdatafile", adios2.Mode.Read)
    while True:
        stepStatus = reader.BeginStep()
        if stepStatus == adios2.StepStatus.OK:
            # inquire for variable
            recvar = io.InquireVariable("np_data")
            if recvar:
                got_exception = 0
                # determine the shape of the data that will be sent
                bufshape = recvar.Shape()
                typ = recvar.Type()
                logging.info("        Incoming variable type is " + typ)
                # allocate buffer for new numpy
                data = np.ones(bufshape, dtype=np.float32)
                logging.info(f"        receive buffer  dtype: {data.dtype!s}")
                try:
                    reader.Get(recvar, data, adios2.Mode.Sync)
                except Exception as e:
                    got_exception = 1
                    logging.info(str(e))
                if got_exception:
                    raise Exception("        Got unexpected exception")
                else:
                    logging.info("        No exception as expected")
            else:
                raise ValueError("InquireVariable failed")
        elif stepStatus == adios2.StepStatus.EndOfStream:
            break
        else:
            raise StopIteration(f"next step failed to initiate {stepStatus!s}")
        reader.EndStep()
    reader.Close()
    logging.info(
        " Reader: finished reading",
    )


if __name__ == "__main__":
    main()



from adios2.adios import Adios

import adios2.bindings as bindings

import unittest
import numpy as np


class TestEngine(unittest.TestCase):
    def test_close(self):
        adios = Adios()
        io = adios.declare_io("BPWriter")
        with io.open("pythontestengine.bp", bindings.Mode.Write):
            pass

    def test_put(self):
        adios = Adios()
        with adios.declare_io("BPWriter") as io:
            pressure = io.define_variable("pressure")
            temps = io.define_variable("temps", np.empty([1], dtype=np.int64))
            with io.open("pythontestengine.bp", bindings.Mode.Write) as engine:
                engine.put(pressure, "35PSI")
                temps_measures = np.array([35, 40, 30, 45], dtype=np.int64)
                engine.put(temps, temps_measures)

    def test_get(self):
        adios = Adios()
        with adios.declare_io("BPWriter") as io:
            pressure = io.define_variable("pressure")
            temps = io.define_variable(
                name="temps",
                content=np.empty([1], dtype=np.int64),
                start=[0],
                shape=[4],
                count=[4],
            )
            with io.open("pythontestengine.bp", bindings.Mode.Write) as engine:
                engine.put(pressure, "35PSI")
                temps_measures = np.array([35, 40, 30, 45], dtype=np.int64)
                engine.put(temps, temps_measures)

        with adios.declare_io("BPReader") as reader:
            with reader.open("pythontestengine.bp", bindings.Mode.Read) as engine:
                engine.begin_step()
                pressure = reader.inquire_variable("pressure")
                temps = reader.inquire_variable("temps")
                pressure_reading = engine.get(pressure)
                temps_reading = np.empty([4], dtype=np.int64)
                engine.get(temps, temps_reading)
                info = engine.blocks_info("temps", 0)
                engine.end_step()
                self.assertEqual(pressure_reading, "35PSI")
                self.assertTrue(np.array_equal(temps_reading, np.array([35, 40, 30, 45])))

                self.assertIsNot(info, None)
                self.assertEqual(info[0]["Start"], "0")
                self.assertEqual(info[0]["Count"], "4")
                self.assertEqual(info[0]["WriterID"], "0")

    def test_steps(self):
        adios = Adios()
        with adios.declare_io("BPWriter") as io:
            pressure = io.define_variable("pressure")
            with io.open("pythontestengine.bp", bindings.Mode.Write) as engine:
                for step in range(0, 10):
                    engine.begin_step()
                    self.assertTrue(engine.between_step_pairs())
                    engine.put(pressure, f"{step}PSI")
                    engine.end_step()
                    self.assertFalse(engine.between_step_pairs())

        with adios.declare_io("BPReader") as reader:
            with reader.open("pythontestengine.bp", bindings.Mode.Read) as engine:
                for i in range(0, engine.steps()):
                    engine.begin_step()
                    pressure = reader.inquire_variable("pressure")
                    pressure_reading = engine.get(pressure)
                    self.assertEqual(pressure_reading, f"{i}PSI")
                    engine.end_step()

    def test_blockinfo(self):
        adios = Adios()
        with adios.declare_io("BPWriter") as io:
            temps = io.define_variable("temps", np.empty([4], dtype=np.int64))
            with io.open("pythontestengine.bp", bindings.Mode.Write) as engine:
                temps_measures = np.array([35, 40, 30, 45], dtype=np.int64)
                engine.put(temps, temps_measures)

        with adios.declare_io("BPReader") as reader:
            with reader.open("pythontestengine.bp", bindings.Mode.Read) as engine:
                engine.begin_step()
                all_blocks = engine.all_blocks_info("temps")
                temps = reader.inquire_variable("temps")
                info = engine.blocks_info("temps", 0)
                engine.end_step()

                self.assertIsNot(info, None)
                self.assertEqual(info[0]["Start"], "0")
                self.assertEqual(info[0]["Count"], "0")
                self.assertEqual(info[0]["WriterID"], "0")
                self.assertEqual(info, all_blocks[0])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import argparse
import atexit
import base64
from datetime import datetime, timedelta, timezone
import dateutil.parser
from github import Github
import json
import os
import re
import subprocess
import sys
import tempfile
import urllib.parse
import urllib.request


class SpackCIBridge(object):

    def __init__(self, gitlab_repo="", gitlab_host="", gitlab_project="", github_project="",
                 disable_status_post=True, sync_draft_prs=False,
                 main_branch=None, prereq_checks=[]):
        self.gitlab_repo = gitlab_repo
        self.github_project = github_project
        github_token = os.environ.get('GITHUB_TOKEN')
        self.github_repo = "https://{0}@github.com/{1}.git".format(github_token, self.github_project)
        self.py_github = Github(github_token)
        self.py_gh_repo = self.py_github.get_repo(self.github_project, lazy=True)

        self.merge_msg_regex = re.compile(r"Merge\s+([^\s]+)\s+into\s+([^\s]+)")
        self.unmergeable_shas = []

        self.post_status = not disable_status_post
        self.sync_draft_prs = sync_draft_prs
        self.main_branch = main_branch
        self.currently_running_sha = None
        self.latest_tested_main_commit = None

        self.prereq_checks = prereq_checks

        dt = datetime.now(timezone.utc) + timedelta(minutes=-60)
        self.time_threshold_brief = urllib.parse.quote_plus(dt.isoformat(timespec="seconds"))

        self.pipeline_api_template = gitlab_host
        self.pipeline_api_template += "/api/v4/projects/"
        self.pipeline_api_template += urllib.parse.quote_plus(gitlab_project)
        self.pipeline_api_template += "/pipelines?ref={0}"

        self.commit_api_template = gitlab_host
        self.commit_api_template += "/api/v4/projects/"
        self.commit_api_template += urllib.parse.quote_plus(gitlab_project)
        self.commit_api_template += "/repository/commits/{0}"

        self.cached_commits = {}

    @atexit.register
    def cleanup():
        """Shutdown ssh-agent upon program termination."""
        if "SSH_AGENT_PID" in os.environ:
            print("    Shutting down ssh-agent({0})".format(os.environ["SSH_AGENT_PID"]))
            subprocess.run(["ssh-agent", "-k"], check=True)

    def setup_ssh(self, ssh_key_base64):
        """Start the ssh agent."""
        print("Starting ssh-agent")
        output = subprocess.run(["ssh-agent", "-s"], check=True, stdout=subprocess.PIPE).stdout

        # Search for PID in output.
        pid_regexp = re.compile(r"SSH_AGENT_PID=([0-9]+)")
        match = pid_regexp.search(output.decode("utf-8"))
        if match is None:
            print("WARNING: could not detect ssh-agent PID.", file=sys.stderr)
            print("ssh-agent will not be killed upon program termination", file=sys.stderr)
        else:
            pid = match.group(1)
            os.environ["SSH_AGENT_PID"] = pid
            self.cleanup_ssh_agent = True

        # Search for socket in output.
        socket_regexp = re.compile(r"SSH_AUTH_SOCK=([^;]+);")
        match = socket_regexp.search(output.decode("utf-8"))
        if match is None:
            print("WARNING: could not detect ssh-agent socket.", file=sys.stderr)
            print("Key will be added to caller's ssh-agent (if any)", file=sys.stderr)
        else:
            socket = match.group(1)
            os.environ["SSH_AUTH_SOCK"] = socket

        # Add the key.
        ssh_key = base64.b64decode(ssh_key_base64)
        ssh_key = ssh_key.replace(b"\r", b"")
        with tempfile.NamedTemporaryFile() as fp:
            fp.write(ssh_key)
            fp.seek(0)
            subprocess.run(["ssh-add", fp.name], check=True)

    def get_commit(self, commit):
        """ Check our cache for a commit on GitHub.
            If we don't have it yet, use the GitHub API to retrieve it."""
        if commit not in self.cached_commits:
            self.cached_commits[commit] = self.py_gh_repo.get_commit(sha=commit)
        return self.cached_commits[commit]

    def list_github_prs(self):
        """ Return two dicts of data about open PRs on GitHub:
            one for all open PRs, and one for open PRs that are not up-to-date on GitLab."""
        pr_dict = {}
        pulls = self.py_gh_repo.get_pulls(state="open")
        print("Rate limit after get_pulls(): {}".format(self.py_github.rate_limiting[0]))
        for pull in pulls:
            backlogged = False
            push = True
            if pull.draft and not self.sync_draft_prs:
                print("Skipping draft PR {0} ({1})".format(pull.number, pull.head.ref))
                backlogged = "draft"
                push = False

            pr_string = "pr{0}_{1}".format(pull.number, pull.head.ref)

            if push and pull.updated_at < datetime.now() + timedelta(minutes=-2880):
                # Skip further analysis of this PR if it hasn't been updated in 48 hours.
                # This helps us avoid wasting our rate limit on PRs with merge conflicts.
                print("Skip pushing stale PR {0}".format(pr_string))
                backlogged = "stale"
                push = False

            if push:
                # Determine if this PR still needs to be pushed to GitLab. This happens in one of two cases:
                # 1) we have never pushed it before
                # 2) we have pushed it before, but the HEAD sha has changed since we pushed it last
                log_args = ["git", "log", "--pretty=%s", "gitlab/{0}".format(pr_string)]
                try:
                    merge_commit_msg = subprocess.run(
                        log_args, check=True, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL).stdout
                    match = self.merge_msg_regex.match(merge_commit_msg.decode("utf-8"))
                    if match and (match.group(1) == pull.head.sha or match.group(2) == pull.head.sha):
                        print("Skip pushing {0} because GitLab already has HEAD {1}".format(pr_string, pull.head.sha))
                        push = False
                except subprocess.CalledProcessError:
                    # This occurs when it's a new PR that hasn't been pushed to GitLab yet.
                    pass

            if push:
                # Check the PRs-to-be-pushed to see if any of them should be considered "backlogged".
                # We currently recognize three types of backlogged PRs:
                # 1) Some required "prerequisite checks" have not yet completed successfully.
                # 2) The PR is based on a version of the "main branch" that has not yet been tested
                # 3) Draft PRs. Handled earlier in this function.
                if not backlogged and self.prereq_checks:
                    checks_desc = "waiting for {} check to succeed"
                    checks_to_verify = self.prereq_checks.copy()
                    pr_check_runs = self.get_commit(pull.head.sha).get_check_runs()
                    for check in pr_check_runs:
                        if check.name in checks_to_verify:
                            checks_to_verify.remove(check.name)
                            if check.conclusion != "success":
                                backlogged = checks_desc.format(check.name)
                                push = False
                                break
                    if not backlogged and checks_to_verify:
                        backlogged = checks_desc.format(checks_to_verify[0])
                        push = False
                    if backlogged:
                        print("Skip pushing {0} because of {1}".format(pr_string, backlogged))

                if not backlogged:
                    if self.main_branch and pull.base.ref == self.main_branch:
                        # Check if we should defer pushing/testing this PR because it is based on "too new" of a commit
                        # of the main branch.
                        tmp_pr_branch = f"temporary_{pr_string}"
                        subprocess.run(["git", "fetch", "github",
                                       f"refs/pull/{pull.number}/head:{tmp_pr_branch}"], check=True)
                        # Get the merge base between this PR and the main branch.
                        try:
                            merge_base_sha = subprocess.run(
                                ["git", "merge-base", tmp_pr_branch, f"github/{self.main_branch}"],
                                check=True, stdout=subprocess.PIPE).stdout.strip()
                        except subprocess.CalledProcessError:
                            print(f"'git merge-base {tmp_pr_branch} github/{self.main_branch}' "
                                  "returned non-zero. Skipping")
                            self.unmergeable_shas.append(pull.head.sha)
                            continue

                        repo_head_sha = subprocess.run(
                            ["git", "rev-parse", tmp_pr_branch],
                            check=True, stdout=subprocess.PIPE).stdout.decode("utf-8").strip()

                        if pull.head.sha != repo_head_sha:
                            # If gh repo and api don't agree on what the head sha is, don't
                            # push.  Instead log an error message and backlog the PR.
                            a_sha, r_sha = pull.head.sha[:7], repo_head_sha[:7]
                            print(f"Skip pushing {pr_string} because api says HEAD is {a_sha}, "
                                  f"while repo says HEAD is {r_sha}")
                            backlogged = f"GitHub HEAD shas out of sync (repo={r_sha}, API={a_sha})"
                            push = False
                        # Check if our PR's merge base is an ancestor of the latest tested main branch commit.
                        elif subprocess.run(
                                ["git", "merge-base", "--is-ancestor", merge_base_sha, self.latest_tested_main_commit]
                                ).returncode == 0:
                            print(f"{tmp_pr_branch}'s merge base IS an ancestor of latest_tested_main "
                                  f"{merge_base_sha} vs. {self.latest_tested_main_commit}")
                            try:
                                subprocess.run(["git", "checkout", self.latest_tested_main_commit], check=True)
                                subprocess.run(["git", "checkout", "-b", pr_string], check=True)
                                commit_msg = f"Merge {pull.head.sha} into {self.latest_tested_main_commit}"
                                subprocess.run(
                                    ["git", "merge", "--no-ff", "-m", commit_msg, tmp_pr_branch],
                                    check=True)
                                print(f"Merge succeeded, ready to push {pr_string} to GitLab for CI pipeline testing")
                            except subprocess.CalledProcessError:
                                print(f"Failed to merge PR {pull.number} ({pull.head.ref}) with latest tested "
                                      f"{self.main_branch} ({self.latest_tested_main_commit}). Skipping")
                                self.unmergeable_shas.append(pull.head.sha)
                                subprocess.run(["git", "merge", "--abort"])
                                backlogged = "merge conflicts with {}".format(self.main_branch)
                                push = False
                                continue
                        else:
                            print(f"Skip pushing {pr_string} because its merge base is NOT an ancestor of "
                                  f"latest_tested_main {merge_base_sha} vs. {self.latest_tested_main_commit}")
                            backlogged = "base"
                            push = False
                    else:
                        # If the --main-branch CLI argument wasn't passed, or if this PR doesn't target that branch,
                        # then we will push the merge commit that was automatically created by GitHub to GitLab
                        # where it will kick off a CI pipeline.
                        try:
                            subprocess.run(["git", "fetch", "github",
                                           f"{pull.merge_commit_sha}:{pr_string}"], check=True)
                        except subprocess.CalledProcessError:
                            print("Failed to locally checkout PR {0} ({1}). Skipping"
                                  .format(pull.number, pull.merge_commit_sha))
                            backlogged = "GitLab failed to checkout this branch"
                            push = False
                            continue

            pr_dict[pr_string] = {
                'base_sha': pull.base.sha,
                'head_sha': pull.head.sha,
                'push': push,
                'backlogged': backlogged,
            }

        def listify_dict(d):
            pr_strings = sorted(d.keys())
            base_shas = [d[s]['base_sha'] for s in pr_strings]
            head_shas = [d[s]['head_sha'] for s in pr_strings]
            b_logged = [d[s]['backlogged'] for s in pr_strings]
            return {
                "pr_strings": pr_strings,
                "base_shas": base_shas,
                "head_shas": head_shas,
                "backlogged": b_logged,
            }
        all_open_prs = listify_dict(pr_dict)
        filtered_pr_dict = {k: v for (k, v) in pr_dict.items() if v['push']}
        filtered_open_prs = listify_dict(filtered_pr_dict)
        print("All Open PRs:")
        for pr_string in all_open_prs['pr_strings']:
            print("    {0}".format(pr_string))
        print("Filtered Open PRs:")
        for pr_string in filtered_open_prs['pr_strings']:
            print("    {0}".format(pr_string))
        print("Rate limit at the end of list_github_prs(): {}".format(self.py_github.rate_limiting[0]))
        return [all_open_prs, filtered_open_prs]

    def list_github_protected_branches(self):
        """ Return a list of protected branch names from GitHub."""
        branches = self.py_gh_repo.get_branches()
        print("Rate limit after get_branches(): {}".format(self.py_github.rate_limiting[0]))
        protected_branches = [br.name for br in branches if br.protected]
        protected_branches = sorted(protected_branches)
        if self.currently_running_sha:
            print("Skip pushing {0} because it already has a pipeline running ({1})"
                  .format(self.main_branch, self.currently_running_sha))
            protected_branches.remove(self.main_branch)
        print("Protected branches:")
        for protected_branch in protected_branches:
            print("    {0}".format(protected_branch))
        return protected_branches

    def list_github_tags(self):
        """ Return a list of tag names from GitHub."""
        tag_list = self.py_gh_repo.get_tags()
        print("Rate limit after get_tags(): {}".format(self.py_github.rate_limiting[0]))
        tags = sorted([tag.name for tag in tag_list])
        print("Tags:")
        for tag in tags:
            print("    {0}".format(tag))
        return tags

    def setup_git_repo(self):
        """Initialize a bare git repository with two remotes:
        one for GitHub and one for GitLab.
        If main_branch was specified, we also fetch that branch from GitHub.
        """
        subprocess.run(["git", "init"], check=True)
        subprocess.run(["git", "config", "user.email", "noreply@spack.io"], check=True)
        subprocess.run(["git", "config", "user.name", "spackbot"], check=True)
        subprocess.run(["git", "config", "advice.detachedHead", "false"], check=True)
        subprocess.run(["git", "remote", "add", "github", self.github_repo], check=True)
        subprocess.run(["git", "remote", "add", "gitlab", self.gitlab_repo], check=True)

        # Shallow fetch from GitLab.
        self.gitlab_shallow_fetch()

        if self.main_branch:
            subprocess.run(["git", "fetch", "github", self.main_branch], check=True)

    def get_gitlab_pr_branches(self):
        """Query GitLab for branches that have already been copied over from GitHub PRs.
        Return the string output of `git branch --remotes --list gitlab/pr*`.
        """
        branch_args = ["git", "branch", "--remotes", "--list", "gitlab/pr*"]
        self.gitlab_pr_output = \
            subprocess.run(branch_args, check=True, stdout=subprocess.PIPE).stdout

    def gitlab_shallow_fetch(self):
        """Perform a shallow fetch from GitLab"""
        fetch_args = ["git", "fetch", "-q", "--depth=1", "gitlab"]
        subprocess.run(fetch_args, check=True, stdout=subprocess.PIPE).stdout

    def get_open_refspecs(self, open_prs):
        """Return a list of refspecs to push given a list of open PRs."""
        print("Building initial lists of refspecs to fetch and push")
        pr_strings = open_prs["pr_strings"]
        base_shas = open_prs["base_shas"]
        backlogged = open_prs["backlogged"]
        open_refspecs = []
        for open_pr, base_sha, backlog in zip(pr_strings, base_shas, backlogged):
            open_refspecs.append("{0}:{0}".format(open_pr))
            print("  pushing {0} (based on {1})".format(open_pr, base_sha))
        return open_refspecs

    def update_refspecs_for_protected_branches(self, protected_branches, open_refspecs, fetch_refspecs):
        """Update our refspecs lists for protected branches from GitHub."""
        for protected_branch in protected_branches:
            fetch_refspecs.append("+refs/heads/{0}:refs/remotes/{0}".format(protected_branch))
            open_refspecs.append("refs/heads/{0}:refs/heads/{0}".format(protected_branch))
        return open_refspecs, fetch_refspecs

    def update_refspecs_for_tags(self, tags, open_refspecs, fetch_refspecs):
        """Update our refspecs lists for tags from GitHub."""
        for tag in tags:
            fetch_refspecs.append("+refs/tags/{0}:refs/tags/{0}".format(tag))
            open_refspecs.append("refs/tags/{0}:refs/tags/{0}".format(tag))
        return open_refspecs, fetch_refspecs

    def fetch_github_branches(self, fetch_refspecs):
        """Perform `git fetch` for a given list of refspecs."""
        print("Fetching GitHub refs for open PRs")
        fetch_args = ["git", "fetch", "-q", "github"] + fetch_refspecs
        subprocess.run(fetch_args, check=True)

    def build_local_branches(self, protected_branches):
        """Create local branches for a list of protected branches."""
        print("Building local branches for protected branches")
        for branch in protected_branches:
            local_branch_name = "{0}".format(branch)
            remote_branch_name = "refs/remotes/{0}".format(branch)
            subprocess.run(["git", "branch", "-q", local_branch_name, remote_branch_name], check=True)

    def make_status_for_pipeline(self, pipeline):
        """Generate POST data to create a GitHub status from a GitLab pipeline
           API response
        """
        post_data = {}
        if "status" not in pipeline:
            return post_data

        if pipeline["status"] == "created":
            post_data["state"] = "pending"
            post_data["description"] = "Pipeline has been created"

        elif pipeline["status"] == "waiting_for_resource":
            post_data["state"] = "pending"
            post_data["description"] = "Pipeline is waiting for resources"

        elif pipeline["status"] == "preparing":
            post_data["state"] = "pending"
            post_data["description"] = "Pipeline is preparing"

        elif pipeline["status"] == "pending":
            post_data["state"] = "pending"
            post_data["description"] = "Pipeline is pending"

        elif pipeline["status"] == "running":
            post_data["state"] = "pending"
            post_data["description"] = "Pipeline is running"

        elif pipeline["status"] == "manual":
            post_data["state"] = "pending"
            post_data["description"] = "Pipeline is running manually"

        elif pipeline["status"] == "scheduled":
            post_data["state"] = "pending"
            post_data["description"] = "Pipeline is scheduled"

        elif pipeline["status"] == "failed":
            post_data["state"] = "error"
            post_data["description"] = "Pipeline failed"

        elif pipeline["status"] == "canceled":
            # Do not post canceled pipeline status to GitHub, it's confusing to our users.
            # This usually happens when a PR gets force-pushed. The next time the sync script runs
            # it will post a status for the newly force-pushed commit.
            return {}

        elif pipeline["status"] == "skipped":
            post_data["state"] = "failure"
            post_data["description"] = "Pipeline was skipped"

        elif pipeline["status"] == "success":
            post_data["state"] = "success"
            post_data["description"] = "Pipeline succeeded"

        post_data["target_url"] = pipeline["web_url"]
        return post_data

    def dedupe_pipelines(self, api_response):
        """Prune pipelines API response to only include the most recent result for each SHA"""
        pipelines = {}
        for response in api_response:
            sha = response['sha']
            if sha not in pipelines:
                pipelines[sha] = response
            else:
                existing_datetime = dateutil.parser.parse(pipelines[sha]['updated_at'])
                current_datetime = dateutil.parser.parse(response['updated_at'])
                if current_datetime > existing_datetime:
                    pipelines[sha] = response
        return pipelines

    def find_pr_sha(self, tested_sha):
        api_url = self.commit_api_template.format(tested_sha)

        try:
            request = urllib.request.Request(api_url)
            if "GITLAB_TOKEN" in os.environ:
                request.add_header("Authorization", "Bearer %s" % os.environ["GITLAB_TOKEN"])
            response = urllib.request.urlopen(request)
        except OSError:
            print('Failed to fetch commit for tested sha {0}'.format(tested_sha))
            return None

        response_data = response.read()

        try:
            tested_commit_info = json.loads(response_data)
        except json.decoder.JSONDecodeError:
            print('Failed to parse response as json ({0})'.format(response_data))
            return None

        if 'title' not in tested_commit_info:
            print('Returned commit object missing "Title" field')
            return None

        merge_commit_msg = tested_commit_info['title']
        m = self.merge_msg_regex.match(merge_commit_msg)

        if m is None:
            print('Failed to find pr_sha in merge commit message')
            return None

        return m.group(1)

    def get_pipelines_for_branch(self, branch, time_threshold=None):
        # Use gitlab's API to get pipeline results for the corresponding ref.
        api_url = self.pipeline_api_template.format(
            urllib.parse.quote_plus(branch)
        )

        # Optionally constrain the query with the provided time_threshold
        if time_threshold:
            api_url = "{0}&updated_after={1}".format(api_url, time_threshold)

        try:
            request = urllib.request.Request(api_url)
            if "GITLAB_TOKEN" in os.environ:
                request.add_header("Authorization", "Bearer %s" % os.environ["GITLAB_TOKEN"])
            response = urllib.request.urlopen(request)
        except OSError as inst:
            print("GitLab API request error accessing {0}".format(api_url))
            print(inst)
            return None
        try:
            pipelines = json.loads(response.read())
        except json.decoder.JSONDecodeError as inst:
            print("Error parsing response to {0}".format(api_url))
            print(inst)
            return None

        return self.dedupe_pipelines(pipelines)

    def post_pipeline_status(self, open_prs, protected_branches):
        print("Rate limit at the beginning of post_pipeline_status(): {}".format(self.py_github.rate_limiting[0]))
        pipeline_branches = []
        backlog_branches = []
        # Split up the open_prs branches into two piles: branches we force-pushed to gitlab
        # and branches we deferred pushing.
        for pr_branch, base_sha, head_sha, backlog in zip(open_prs["pr_strings"],
                                                          open_prs["base_shas"],
                                                          open_prs["head_shas"],
                                                          open_prs["backlogged"]):
            if not backlog:
                pipeline_branches.append(pr_branch)
            else:
                backlog_branches.append((pr_branch, head_sha, backlog))

        pipeline_branches.extend(protected_branches)

        print('Querying pipelines to post status for:')
        for branch in pipeline_branches:
            # Post status to GitHub for each pipeline found.
            pipelines = self.get_pipelines_for_branch(branch, self.time_threshold_brief)
            if not pipelines:
                continue
            for sha, pipeline in pipelines.items():
                post_data = self.make_status_for_pipeline(pipeline)
                if not post_data:
                    continue
                # TODO: associate shas with protected branches, so we do not have to
                # hit an endpoint here, but just use the sha we already know just like
                # we do below for backlogged PR statuses.
                pr_sha = self.find_pr_sha(sha)
                if not pr_sha:
                    print('Could not find github PR sha for tested commit: {0}'.format(sha))
                    print('Using tested commit to post status')
                    pr_sha = sha
                self.create_status_for_commit(pr_sha,
                                              branch,
                                              post_data["state"],
                                              post_data["target_url"],
                                              post_data["description"])

        # Post a status of pending/backlogged for branches we deferred pushing
        print("Posting backlogged status to the following:")
        base_backlog_desc = \
            "This branch's merge-base with {} is newer than the latest commit tested by GitLab".format(self.main_branch)
        for branch, head_sha, reason in backlog_branches:
            if reason == "stale":
                print("Skip posting status for {} because it has not been updated recently".format(branch))
                continue
            elif reason == "base":
                desc = base_backlog_desc
                url = "https://github.com/spack/spack-infrastructure/blob/main/docs/deferred_pipelines.md"
            elif reason == "draft":
                desc = "GitLab CI is disabled for draft PRs"
                url = ""
            else:
                desc = reason
                url = ""
            self.create_status_for_commit(head_sha, branch, "pending", url, desc)

        # Post errors to any PRs that we couldn't merge to latest_tested_main_commit.
        print('Posting unmergeable status to the following:')
        for sha in self.unmergeable_shas:
            print('  {0}'.format(sha))
            self.create_status_for_commit(sha, "", "error", "", f"PR could not be merged with {self.main_branch}")
        print("Rate limit at the end of post_pipeline_status(): {}".format(self.py_github.rate_limiting[0]))

    def create_status_for_commit(self, sha, branch, state, target_url, description):
        context = "OLCF Ascent (Summit)"
        commit = self.get_commit(sha)
        existing_statuses = commit.get_combined_status()
        for status in existing_statuses.statuses:
            if (status.context == context and
                    status.state == state and
                    status.description == description and
                    status.target_url == target_url):
                print("Not posting duplicate status to {} / {}".format(branch, sha))
                return
        try:
            status_response = self.get_commit(sha).create_status(
                state=state,
                target_url=target_url,
                description=description,
                context=context
            )
            if status_response.state != state:
                print("Expected CommitStatus state {0}, got {1}".format(
                    state, status_response.state))
        except Exception as e_inst:
            print('Caught exception posting status for {0}/{1}'.format(branch, sha))
            print(e_inst)
        print("  {0} -> {1}".format(branch, sha))

    def sync(self):
        """Synchronize pull requests from GitHub as branches on GitLab."""

        print("Initial rate limit: {}".format(self.py_github.rate_limiting[0]))
        reset_time = datetime.utcfromtimestamp(self.py_github.rate_limiting_resettime).strftime('%Y-%m-%d %H:%M:%S')
        print("Rate limit will refresh at: {} UTC".format(reset_time))

        # Setup SSH command for communicating with GitLab.
        os.environ["GIT_SSH_COMMAND"] = "ssh -F /dev/null -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no"

        # Work inside a temporary directory that will be deleted when this script terminates.
        with tempfile.TemporaryDirectory() as tmpdirname:
            os.chdir(tmpdirname)

            # Setup the local repo with two remotes.
            self.setup_git_repo()

            if self.main_branch:
                # Find the currently running main branch pipeline, if any, and get the sha.
                # Also get the latest commit on the main branch that has a completed pipeline.
                main_branch_pipelines = self.get_pipelines_for_branch(self.main_branch)
                for sha, pipeline in main_branch_pipelines.items():
                    if self.latest_tested_main_commit is None and \
                            (pipeline['status'] == "success" or pipeline['status'] == "failed"):
                        self.latest_tested_main_commit = sha

                    if self.currently_running_sha is None and pipeline['status'] == "running":
                        self.currently_running_sha = sha

                    if self.latest_tested_main_commit and self.currently_running_sha:
                        break

            print("Latest completed {0} pipeline: {1}".format(self.main_branch, self.latest_tested_main_commit))
            print("Currently running {0} pipeline: {1}".format(self.main_branch, self.currently_running_sha))

            # Retrieve open PRs from GitHub.
            all_open_prs, open_prs = self.list_github_prs()

            # Get protected branches on GitHub.
            protected_branches = self.list_github_protected_branches()

            # Get tags on GitHub.
            tags = self.list_github_tags()

            # Get refspecs for open PRs and protected branches.
            open_refspecs = self.get_open_refspecs(open_prs)
            fetch_refspecs = []
            self.update_refspecs_for_protected_branches(protected_branches, open_refspecs, fetch_refspecs)
            self.update_refspecs_for_tags(tags, open_refspecs, fetch_refspecs)

            # Sync open GitHub PRs and protected branches to GitLab.
            self.fetch_github_branches(fetch_refspecs)
            self.build_local_branches(protected_branches)
            if open_refspecs:
                print("Syncing to GitLab")
                push_args = ["git", "push", "--porcelain", "-f", "gitlab"] + open_refspecs
                subprocess.run(push_args, check=True)

            # Post pipeline status to GitHub for each open PR, if enabled
            if self.post_status:
                print('Posting pipeline status for open PRs and protected branches')
                self.post_pipeline_status(all_open_prs, protected_branches)


if __name__ == "__main__":
    # Parse command-line arguments.
    parser = argparse.ArgumentParser(description="Sync GitHub PRs to GitLab")
    parser.add_argument("github_project", help="GitHub project (org/repo or user/repo)")
    parser.add_argument("gitlab_repo", help="Full clone URL for GitLab")
    parser.add_argument("gitlab_host", help="GitLab web host")
    parser.add_argument("gitlab_project", help="GitLab project (org/repo or user/repo)")
    parser.add_argument("--disable-status-post", action="store_true", default=False,
                        help="Do not post pipeline status to each GitHub PR")
    parser.add_argument("--sync-draft-prs", action="store_true", default=False,
                        help="Copy draft PRs from GitHub to GitLab")
    parser.add_argument("--pr-mirror-bucket", default=None,
                        help="Delete mirrors for closed PRs from the specified S3 bucket")
    parser.add_argument("--main-branch", default=None,
                        help="""If provided, we check if there is a currently running
pipeline for this branch. If so, we defer syncing any subsequent commits in an effort
to not interrupt this pipeline. We also defer pushing any PR branches that are based
on a commit of the main branch that is newer than the latest commit tested by GitLab.""")
    parser.add_argument("--prereq-check", nargs="+", default=False,
                        help="Only push branches that have already passed this GitHub check")

    args = parser.parse_args()

    ssh_key_base64 = os.getenv("GITLAB_SSH_KEY_BASE64")
    if ssh_key_base64 is None:
        raise Exception("GITLAB_SSH_KEY_BASE64 environment is not set")

    if "GITHUB_TOKEN" not in os.environ:
        raise Exception("GITHUB_TOKEN environment is not set")

    bridge = SpackCIBridge(gitlab_repo=args.gitlab_repo,
                           gitlab_host=args.gitlab_host,
                           gitlab_project=args.gitlab_project,
                           github_project=args.github_project,
                           disable_status_post=args.disable_status_post,
                           sync_draft_prs=args.sync_draft_prs,
                           main_branch=args.main_branch,
                           prereq_checks=args.prereq_check)
    bridge.setup_ssh(ssh_key_base64)
    bridge.sync()



#!/usr/bin/env python3

import sys
import argparse
import urllib.request
import json

parser = argparse.ArgumentParser(description="List open PRs")
parser.add_argument("-t", "--token", help="GitHub access token")
group = parser.add_mutually_exclusive_group()
group.add_argument("-o", "--open", help="List open PRs",
                   action="store_true", default=False)
group.add_argument("-c", "--closed", help="List closed PRs",
                   action="store_true", default=False)
group.add_argument("-a", "--all", help="List all PRs",
                   action="store_true", default=False)
parser.add_argument("repo", help="GitHub repo (org/repo or user/repo)")
args = parser.parse_args()

if args.open:
    state = "open"
elif args.closed:
    state = "closed"
elif args.all:
    state = "all"
else:
    state = "open"

try:
    request = urllib.request.Request(
        'https://api.github.com/repos/%s/pulls?state=%s' % (args.repo, state))
    if args.token:
        request.add_header('Authorization', 'token %s' % args.token)
    response = urllib.request.urlopen(request)
except OSError:
    sys.exit(1)

prs = json.loads(response.read())

for pr in prs:
    print("%d %s" % (pr['number'], pr['head']['ref']))



#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# Test reader in Python that should have the same output
# as the test1.m Matlab reader
#
#
#  Created on: Jan 28, 2019
#      Author: Norbert Podhorszki, pnorbert@ornl.gov
#

import adios2

# ADIOS2 high-level API for Reading
fr = adios2.open("test1.bp", "r")

inNRows = fr.read("nrows")
print("# of rows = {0}".format(inNRows[0]))

inNCols = fr.read("ncols")
print("# of cols = {0}".format(inNCols[0]))

inNote = fr.readstring("note")
print("Note = {0}".format(inNote))

inTemperatures = fr.read("temperature2D")
print("temperature2d array size = " + str(inTemperatures.size))

for row in inTemperatures:
    print(''.join(['{:7}'.format(item) for item in row]))

fr.close()



import numpy as np  # pylint: disable=import-error


def Locate(rank, nproc, datasize):
    extra = 0
    if rank == nproc - 1:
        extra = datasize % nproc
    num = datasize // nproc
    start = num * rank
    size = num + extra
    return start, size


class MPISetup(object):
    readargs = []
    size = 1
    rank = {"app": 0, "x": 0, "y": 0}

    def __init__(self, args, appID):

        self.nx = args.nx
        self.ny = args.ny
        self.nz = args.nz

        if not args.nompi:

            from mpi4py import MPI  # pylint: disable=import-error

            self.comm_app = MPI.COMM_WORLD.Split(appID, MPI.COMM_WORLD.Get_rank())
            self.size = self.comm_app.Get_size()
            self.rank["app"] = self.comm_app.Get_rank()
            if self.nx * self.ny * self.nz == 1:
                self.nx = self.size
            if self.size != (self.nx * self.ny * self.nz):
                raise ValueError("nx * ny * nz != num processes")

            if (self.ny > 1) and (self.nx > 1) and (self.nz > 1):
                comm_x = self.comm_app.Split(
                    self.rank["app"] % self.nx, self.rank["app"]
                )
            else:
                comm_x = self.comm_app.Split(
                    self.rank["app"] / self.nx, self.rank["app"]
                )
            comm_y = self.comm_app.Split(self.rank["app"] / self.ny, self.rank["app"])
            comm_z = self.comm_app.Split(self.rank["app"] / self.nz, self.rank["app"])

            self.rank["x"] = comm_x.Get_rank()
            self.rank["y"] = comm_y.Get_rank()
            self.rank["z"] = comm_z.Get_rank()

            self.readargs.append(self.comm_app)

        else:
            if self.nx != 1:
                raise ValueError("nx must = 1 without MPI")
            if self.ny != 1:
                raise ValueError("ny must = 1 without MPI")

    #            self.readargs.extend([args.xmlfile, "heat"])

    def Partition_3D_3D(self, fp, args):
        datashape = np.zeros(3, dtype=np.int64)
        start = np.zeros(3, dtype=np.int64)
        size = np.zeros(3, dtype=np.int64)

        # print("dir fp {0}".format(dir(fp)))
        var = fp.available_variables()
        #        print("var {0}".format(var))
        #        print('key')
        #        print('type {0}'.format(type(var)))
        # print("args varname {0}".format(args.varname))
        # print('var keys')
        #        print(var.keys())
        dshape = var[args.varname]["Shape"].split(",")
        for i in range(len(dshape)):
            datashape[i] = int(dshape[i])

        start[0], size[0] = Locate(self.rank["y"], self.ny, datashape[0])
        start[1], size[1] = Locate(self.rank["x"], self.nx, datashape[1])
        start[0], size[0] = Locate(self.rank["y"], self.ny, datashape[0])

        return start, size, datashape

    def Partition_3D_1D(self, fp, args):
        datashape = np.zeros(3, dtype=np.int64)
        start = np.zeros(3, dtype=np.int64)
        size = np.zeros(3, dtype=np.int64)

        # print("dir fp {0}".format(dir(fp)))
        var = fp.availablevariables()
        #        print("var {0}".format(var))
        #        print('key')
        #        print('type {0}'.format(type(var)))
        # print("args varname {0}".format(args.varname))
        # print('var keys')
        #        print(var.keys())
        dshape = var[args.varname]["Shape"].split(",")
        for i in range(len(dshape)):
            datashape[i] = int(dshape[i])

        start[0], size[0] = Locate(self.rank["app"], self.size, datashape[0])
        start[1], size[1] = (0, datashape[1])
        start[2], size[2] = (0, datashape[2])

        return start, size, datashape

    def Partition_2D_1D(self, fp, args):
        datashape = np.zeros(2, dtype=np.int64)
        start = np.zeros(2, dtype=np.int64)
        size = np.zeros(2, dtype=np.int64)

        var = fp.availablevariables()
        dshape = var[args.varname]["Shape"].split(",")
        for i in range(len(dshape)):
            datashape[i] = int(dshape[i])

        start[0], size[0] = Locate(self.rank["app"], self.size, datashape[0])
        start[1], size[1] = (0, datashape[1])

        return start, size, datashape



from adios2.adios import Adios

import adios2.bindings as bindings

import unittest
import numpy as np


class Test_attribute(unittest.TestCase):
    def test_create_write(self):
        adios = Adios()
        with adios.declare_io("BPWriter") as io:
            ts = io.define_attribute("timestamp", "20231122")
            self.assertEqual(ts.name(), "timestamp")
            self.assertEqual(ts.data_string(), "20231122")

    def test_create_reader(self):
        adios = Adios()
        with adios.declare_io("BPWriter") as io:
            ts = io.define_attribute("timestamp", "20231122")
            self.assertEqual(ts.name(), "timestamp")
            self.assertEqual(ts.data_string(), "20231122")

    def test_create_write_ndarray(self):
        adios = Adios()
        with adios.declare_io("BPWriter") as io:
            arr = np.array([2023, 11, 22])
            ts = io.define_attribute("timestamp", arr)
            self.assertEqual(ts.name(), "timestamp")
            self.assertTrue(np.array_equal(ts.data(), [2023, 11, 22]))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
from adios2 import Adios, Stream  # pylint: disable=import-error
import argparse
import numpy as np  # pylint: disable=import-error
import matplotlib.pyplot as plt  # pylint: disable=import-error
import matplotlib.gridspec as gridspec  # pylint: disable=import-error
import decomp  # pylint: disable=import-error


def SetupArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--instream", "-i", help="Name of the input stream", required=True
    )
    parser.add_argument(
        "--outfile", "-o", help="Name of the output file", default="screen"
    )
    parser.add_argument("--varname", "-v", help="Name of variable read", default="U")
    parser.add_argument(
        "--nompi", "-nompi", help="ADIOS was installed without MPI", action="store_true"
    )
    parser.add_argument(
        "--displaysec",
        "-dsec",
        help="Float representing gap between plot window refresh",
        default=0.2,
    )
    args = parser.parse_args()

    args.displaysec = float(args.displaysec)
    args.nx = 1
    args.ny = 1
    args.nz = 1

    return args


def PlotPDF(pdf, bins, args, start, count, step, fontsize):
    # Plotting part
    displaysec = args.displaysec
    gs = gridspec.GridSpec(1, 1)
    fig = plt.figure(1, figsize=(8, 8))
    ax = fig.add_subplot(gs[0, 0])
    localSliceIdx = count[0] // 2
    globalSliceIdx = start[0] + localSliceIdx

    ax.plot(bins, pdf[localSliceIdx], "r*-")

    ax.set_title(
        "{0}, slice {1}, step {2}".format(args.varname, globalSliceIdx, step),
        fontsize=fontsize,
    )
    ax.set_xlabel("bins", fontsize=fontsize)
    ax.set_ylabel("count", fontsize=fontsize)
    plt.tick_params(labelsize=fontsize - 4)
    plt.ion()
    if args.outfile == "screen":
        plt.show()
        plt.pause(displaysec)
    else:
        imgfile = (
            args.outfile + "{0:0>5}".format(step) + "_" + str(globalSliceIdx) + ".png"
        )
        fig.savefig(imgfile)

    plt.clf()


def read_data(args, fr, start_coord, size_dims):
    var1 = args.varname
    data = fr.read(var1, start_coord, size_dims)
    data = np.squeeze(data)
    return data


if __name__ == "__main__":
    # fontsize on plot
    fontsize = 24

    args = SetupArgs()
    # print(args)

    # Setup up 2D communicators if MPI is installed
    mpi = decomp.MPISetup(args, 4)
    myrank = mpi.rank["app"]

    # Read the data from this object
    adios = Adios("adios2.xml", mpi.comm_app)
    io = adios.declare_io("PDFAnalysisOutput")
    fr = Stream(io, args.instream, "r", mpi.comm_app)

    # Read through the steps, one at a time
    plot_step = 0
    for fr_step in fr.steps():
        cur_step = fr_step.current_step()
        vars_info = fr_step.available_variables()
        # print (vars_info)
        pdfvar = args.varname + "/pdf"
        binvar = args.varname + "/bins"
        shape2_str = vars_info[pdfvar]["Shape"].split(",")
        shape2 = list(map(int, shape2_str))

        # Equally partition the PDF arrays among readers
        s0, c0 = decomp.Locate(myrank, mpi.size, shape2[0])
        start = [s0, 0]
        count = [c0, shape2[1]]
        start_bins = [0]
        count_bins = [shape2[1]]

        # print("Rank {0} reads {1}  slices from offset {2}".format(myrank, count[0], start[0]))

        pdf = fr_step.read(pdfvar, start, count)
        bins = fr_step.read(binvar, start_bins, count_bins)
        sim_step = fr_step.read("step")

        if myrank == 0:
            print(
                "PDF Plot step {0} processing analysis step {1} simulation step {2}".format(
                    plot_step, cur_step, sim_step
                ),
                flush=True,
            )
        # if cur_step == 0:
        #     print("Variable" + pdfvar + " shape is {" + vars_info[pdfvar]["Shape"]+"}")

        PlotPDF(pdf, bins, args, start, count, sim_step, fontsize)
        plot_step = plot_step + 1
        if not args.nompi:
            mpi.comm_app.Barrier()

    fr.close()



from mpi4py import MPI
import numpy as np
from adios2 import Stream, Adios, bindings

# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# ADIOS MPI Communicator
adios = Adios(comm)

# ADIOS IO
io = adios.declare_io("myIO")
io.set_engine("Sst")

with Stream(io, "helloSst", "r", comm) as stream:
    for _ in stream.steps():
        var = stream.inquire_variable("bpFloats")
        shape = var.shape()
        count = int(shape[0] / size)
        start = count * rank
        if rank == size - 1:
            count += shape[0] % size
        floatArray = stream.read("bpFloats", [start], [count])
        currentStep = stream.current_step()
        loopStep = stream.loop_index()
        print(
            "Rank=",
            rank,
            "loop index =",
            loopStep,
            "stream step =",
            currentStep,
            "data =",
            floatArray,
            flush=True,
        )



from mpi4py import MPI
import numpy as np
import adios2.bindings as adios2

# MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# ADIOS MPI Communicator
adios = adios2.ADIOS(comm)

# ADIOS IO
sstIO = adios.DeclareIO("myIO")
sstIO.SetEngine("Sst")


sstReader = sstIO.Open("helloSst", adios2.Mode.Read)
loopStep = 0
while True:
    status = sstReader.BeginStep()
    if not status == adios2.StepStatus.OK:
        break
    var = sstIO.InquireVariable("bpFloats")
    shape = var.Shape()
    count = int(shape[0] / size)
    start = count * rank
    if rank == size - 1:
        count += shape[0] % size
    var.SetSelection([[start], [count]])
    # note: we need to use np.float32 to be compatible with data from C++ writer
    # using "float" works in Python only but leads to type mismatch with C++
    floatArray = np.zeros(count, dtype=np.float32)
    currentStep = sstReader.CurrentStep()
    sstReader.Get(var, floatArray)
    sstReader.EndStep()

    print(
        "Rank=",
        rank,
        "loop index =",
        loopStep,
        "stream step =",
        currentStep,
        "data =",
        floatArray,
        flush=True,
    )
    loopStep = loopStep + 1

print(f"Exited loop with StepStatus = {status}")
sstReader.Close()



#!/usr/bin/env python
#
# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestDataMan1D.py: test for 1D data transfer by reading in Python
#  Created on: March 3, 2023
#      Author: Dmitry Ganyushin ganyushindi@ornl.gov
from multiprocessing import Process
import unittest
import numpy as np
import adios2


class TestDataMan1D(unittest.TestCase):

    def setUp(self):
        self.conf = {
            "IPAddress": "127.0.0.1",
            "Port": "12306",
            "Timeout": "5",
            "TransportMode": "reliable",
            "RendezvousReaderCount": "1",
        }
        self.Nx = 10
        self.fill_value = 1.0
        self.shape = [1, self.Nx]

    def test_run(self):

        s = Process(target=self.thread_send)
        r = Process(target=self.thread_receive)

        s.start()
        r.start()

        r.join()
        s.join()

    def thread_send(self):
        data = np.full(shape=self.shape, fill_value=self.fill_value)
        shape = data.shape
        count = shape
        start = (0,) * len(shape)

        adios_io = adios2.ADIOS()
        wan = adios_io.DeclareIO("Server")
        wan.SetEngine("Dataman")

        wan.SetParameters(self.conf)
        writer = wan.Open("testdata", adios2.Mode.Write)
        sendbuffer = wan.DefineVariable("np_data", data, shape,
                                        start, count, adios2.ConstantDims)
        self.assertIsNotNone(sendbuffer)
        if sendbuffer:
            writer.BeginStep()
            writer.Put(sendbuffer, data, adios2.Mode.Deferred)
            writer.EndStep()
        else:
            raise ValueError("DefineVariable failed")

        writer.Close()

    def thread_receive(self):
        data = np.zeros(shape=self.shape)
        adios_io = adios2.ADIOS()
        wan = adios_io.DeclareIO("Client")
        wan.SetEngine("Dataman")
        wan.SetParameters(self.conf)
        reader = wan.Open("testdata", adios2.Mode.Read)
        while True:
            stepStatus = reader.BeginStep()
            if stepStatus == adios2.StepStatus.OK:
                recvar = wan.InquireVariable("np_data")
                self.assertIsNotNone(recvar)
                bufshape = recvar.Shape()
                self.assertTrue(bufshape[0] == 1)
                self.assertTrue(bufshape[1] == self.Nx)
                reader.Get(recvar, data, adios2.Mode.Sync)

            elif stepStatus == adios2.StepStatus.EndOfStream:
                break
            else:
                raise StopIteration()
            reader.EndStep()
        reader.Close()
        self.assertTrue(all([data[0][i] == self.fill_value for i
                             in range(len(data))]))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# Distributed under the OSI-approved Apache License, Version 2.0.  See
# accompanying file Copyright.txt for details.
#
# TestBPWriteTypesHighLevelAPILocal.py: test Python numpy types in ADIOS2 File
#                      Write/Read High-Level API for Local Arrays
#  Created on: March 12, 2018
#      Author: William F Godoy godoywf@ornl.gov

from adios2NPTypes import SmallTestData
from mpi4py import MPI
from adios2 import Stream


def check_array(np1, np2, hint):
    if not (np1 == np2).all():
        print("InData: " + str(np1))
        print("Data: " + str(np2))
        raise ValueError("Array read failed " + str(hint))


comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

# Test data
data = SmallTestData(rank)

# Writer
with Stream("types_np_local.bp", "w", comm=comm) as s:
    for _ in s.steps(5):
        step = s.current_step()
        nx = data.Nx - rank - step
        if nx < 1:
            nx = data.Nx
        shape = []
        start = []
        count = [nx]

        if step == 0 and size > 1 and rank == size - 1:
            continue

        data.update(rank, step, size)
        s.write("varI8", data.i8, shape, start, count)
        s.write("varI16", data.i16, shape, start, count)
        v = s.inquire_variable("varI16")
        print(
            f"step {step} rank {rank} nx {nx} count = {v.count()} data.I16 = {data.i16[:nx]}",
            flush=True,
        )

        s.write("varI32", data.i32, shape, start, count)
        s.write("varI64", data.i64, shape, start, count)
        s.write("varU8", data.u8, shape, start, count)
        s.write("varU16", data.u16, shape, start, count)
        s.write("varU32", data.u32, shape, start, count)
        s.write("varU64", data.u64, shape, start, count)
        s.write("varR32", data.r32, shape, start, count)
        s.write("varR64", data.r64, shape, start, count)
        s.write("an_int_list", data.int_list, shape, start, count)
        s.write("a_float_list", data.float_list, shape, start, count)
        s.write("a_complex_list", data.complex_list, shape, start, count)
# Reader
data = SmallTestData(rank)

if rank == 0:
    print("====================================", flush=True)
    print("Read data block by block", flush=True)
comm.Barrier()

with Stream("types_np_local.bp", "r", comm=comm) as s:
    for fr_step in s.steps():
        step = fr_step.current_step()
        nx = data.Nx - rank - step
        if nx < 1:
            nx = data.Nx

        if step == 0 and size > 1 and rank == size - 1:
            continue

        data.update(rank, step, size)
        indataI8 = fr_step.read("varI8", block_id=rank)
        indataI16 = fr_step.read("varI16", block_id=rank)
        indataI32 = fr_step.read("varI32", block_id=rank)
        indataI64 = fr_step.read("varI64", block_id=rank)
        indataU8 = fr_step.read("varU8", block_id=rank)
        indataU16 = fr_step.read("varU16", block_id=rank)
        indataU32 = fr_step.read("varU32", block_id=rank)
        indataU64 = fr_step.read("varU64", block_id=rank)
        indataR32 = fr_step.read("varR32", block_id=rank)
        indataR64 = fr_step.read("varR64", block_id=rank)

        in_int_list = fr_step.read("an_int_list", block_id=rank)

        print(
            f"step {step} rank {rank} nx {nx} I16={indataI16} data.I16 = {data.i16[:nx]}",
            flush=True,
        )

        check_array(indataI8, data.i8[:nx], "i8")
        check_array(indataI16, data.i16[:nx], "i16")
        check_array(indataI32, data.i32[:nx], "i32")
        check_array(indataI64, data.i64[:nx], "i64")
        check_array(indataU8, data.u8[:nx], "u8")
        check_array(indataU16, data.u16[:nx], "u16")
        check_array(indataU32, data.u32[:nx], "u32")
        check_array(indataU64, data.u64[:nx], "u64")
        check_array(indataR32, data.r32[:nx], "r32")
        check_array(indataR64, data.r64[:nx], "r64")

if rank == 0:
    print("====================================", flush=True)
    print("Test blocks_info", flush=True)
    with Stream("types_np_local.bp", "rra") as f:
        v = f.inquire_variable("varI16")
        for step in range(v.steps()):
            bis = f.engine.blocks_info("varI16", step)
            expected_num_blocks = size
            if step == 0 and size > 1:
                expected_num_blocks = size - 1
            print(
                f"step {step} rank {rank} number of blocks expected = {expected_num_blocks}"
                f", reported = {len(bis)}",
                flush=True,
            )
            if len(bis) != expected_num_blocks:
                raise ValueError(
                    f"Expected number of blocks in step {step} "
                    f"is {expected_num_blocks} but we found "
                    f"only {len(bis)} blocks."
                )

            for r in range(size):
                nx = data.Nx - r - step
                if nx < 1:
                    nx = data.Nx
                if step == 0 and size > 1 and r == size - 1:
                    continue
                cnt = int(bis[r]["Count"])
                print(f"    block {r} size expected = {nx}, reported = {cnt}", flush=True)
                if nx != cnt:
                    raise ValueError(
                        f"Expected size of block in step {step}, rank {r} "
                        f"is {nx} but we got reported to have "
                        f"{cnt} elements."
                    )


comm.Barrier()
